// MgNetUDP.cpp
#include "MgNetSys.h"
#include "MgNetLib.h"

class MgNetUDPSocket
{
public:
	int ready;
	SOCKET channel;
	MgNetIPAddress address;
	
	class MgNetUDPChannel
	{
	public:
		int numbound;
		MgNetIPAddress address[MGNET_MAX_UDPADDRESSES];
	};
	
	MgNetUDPChannel binding[MGNET_MAX_UDPCHANNELS];
};


/* Allocate/free a single UDP packet 'size' bytes long.
   The new packet is returned, or NULL if the function ran out of memory.
 */
MgNetUDPPacket *MgNet_AllocPacket(int size)
{
	MgNetUDPPacket *packet;
	int error;


	error = 1;
	packet = (MgNetUDPPacket *)malloc(sizeof(*packet));
	if ( packet != NULL ) {
		packet->maxlen = size;
		packet->data = (Uint8 *)malloc(size);
		if ( packet->data != NULL ) {
			error = 0;
		}
	}
	if ( error ) {
		MgNet_FreePacket(packet);
		packet = NULL;
	}
	return(packet);
}
int MgNet_ResizePacket(MgNetUDPPacket *packet, int newsize)
{
	Uint8 *newdata;

	newdata = (Uint8 *)malloc(newsize);
	if ( newdata != NULL ) {
		free(packet->data);
		packet->data = newdata;
		packet->maxlen = newsize;
	}
	return(packet->maxlen);
}
 void MgNet_FreePacket(MgNetUDPPacket *packet)
{
	if ( packet ) {
		if ( packet->data )
			free(packet->data);
		free(packet);
	}
}

/* Allocate/Free a UDP packet vector (array of packets) of 'howmany' packets,
   each 'size' bytes long.
   A pointer to the packet array is returned, or NULL if the function ran out
   of memory.
 */
MgNetUDPPacket **MgNet_AllocPacketV(int howmany, int size)
{
	MgNetUDPPacket **packetV;

	packetV = (MgNetUDPPacket **)malloc((howmany+1)*sizeof(*packetV));
	if ( packetV != NULL ) {
		int i;
		for ( i=0; i<howmany; ++i ) {
			packetV[i] = MgNet_AllocPacket(size);
			if ( packetV[i] == NULL ) {
				break;
			}
		}
		packetV[i] = NULL;

		if ( i != howmany ) {
			MgNet_FreePacketV(packetV);
			packetV = NULL;
		}
	}
	return(packetV);
}
void MgNet_FreePacketV(MgNetUDPPacket **packetV)
{
	if ( packetV ) {
		int i;
		for ( i=0; packetV[i]; ++i ) {
			MgNet_FreePacket(packetV[i]);
		}
		free(packetV);
	}
}

/* Since the UNIX/Win32/BeOS code is so different from MacOS,
   we'll just have two completely different sections here.
*/

/* Open a UDP network socket
   If 'port' is non-zero, the UDP socket is bound to a fixed local port.
*/
 MgNetUDPSocket * MgNet_UDP_Open(Uint16 port)
{
	MgNetUDPSocket * sock;

	/* Allocate a UDP socket structure */
	sock = (MgNetUDPSocket *)malloc(sizeof(*sock));
	if ( sock == NULL ) {
		MgNet_SetError("Out of memory");
		goto error_return;
	}
	memset(sock, 0, sizeof(*sock));
	
	/* Open the socket */
#ifdef MACOS_OPENTRANSPORT
	{
	OSStatus status;

#if ! TARGET_API_MAC_CARBON
		sock->channel = OTOpenEndpoint(OTCreateConfiguration(kUDPName),0, nil, &status);
#else
		sock->channel = OTOpenEndpointInContext(OTCreateConfiguration(kUDPName),0, nil, &status, nil );
#endif
		if (status != noErr)
		{
			MgNet_SetError("Couldn't create socket, OTOpenEndpoint() = %d",(int) status);
			goto error_return;
		}
	}
#else
	sock->channel = socket(AF_INET, SOCK_DGRAM, 0);
#endif /* MACOS_OPENTRANSPORT */

	if ( sock->channel == INVALID_SOCKET )
	{
		MgNet_SetError("Couldn't create socket");
		goto error_return;
	}

#ifdef MACOS_OPENTRANSPORT
	{
	InetAddress wanted;
	InetAddress assigned;
	TBind wanted_addr;
	TBind assigned_addr;
	OSStatus status;
	InetInterfaceInfo info;

		memset(&assigned_addr, 0, sizeof(assigned_addr));
		assigned_addr.addr.maxlen = sizeof(assigned);
		assigned_addr.addr.len = sizeof(assigned);
		assigned_addr.addr.buf = (UInt8 *) &assigned;
		
		memset(&wanted_addr, 0, sizeof(wanted_addr));
		wanted_addr.addr.maxlen = sizeof(wanted);
		wanted_addr.addr.len = sizeof(wanted);
		wanted_addr.addr.buf = (UInt8 *) &wanted;
		
		status = OTInetGetInterfaceInfo(&info, kDefaultInetInterface);
		
		if (status != noErr)
		{
			MgNet_SetError("Couldn't get interface info, OTInetGetInterfaceInfo() = %d",(int) status);
			goto error_return;
		}
		
		OTInitInetAddress(&wanted, port, info.fAddress);

		if (port == 0)
		{
			status = OTBind(sock->channel, nil, &assigned_addr);
		}
		
		else
		{
			status = OTBind(sock->channel, &wanted_addr, &assigned_addr);
		}
			
		if (status != noErr)
		{
			MgNet_SetError("Couldn't bind to local port, OTBind() = %d",(int) status);
			goto error_return;
		}

		sock->address.host = assigned.fHost;
		sock->address.port = assigned.fPort;
	}
#else
	/* Bind locally, if appropriate */
	if ( port ) {
		struct sockaddr_in sock_addr;
		memset(&sock_addr, 0, sizeof(sock_addr));
		sock_addr.sin_family = AF_INET;
		sock_addr.sin_addr.s_addr = INADDR_ANY;
		sock_addr.sin_port = SDL_SwapBE16(port);

		/* Bind the socket for listening */
		if ( bind(sock->channel, (struct sockaddr *)&sock_addr,
				sizeof(sock_addr)) == SOCKET_ERROR ) {
			MgNet_SetError("Couldn't bind to local port");
			goto error_return;
		}
		/* Fill in the channel host address */
		sock->address.host = sock_addr.sin_addr.s_addr;
		sock->address.port = sock_addr.sin_port;
	}
#endif /* MACOS_OPENTRANSPORT */

	/* The socket is ready */
	
	return(sock);

error_return:
	MgNet_UDP_Close(sock);
	
	return(NULL);
}

/* Verify that the channel is in the valid range */
static int ValidChannel(int channel)
{
	if ( (channel < 0) || (channel >= MGNET_MAX_UDPCHANNELS) ) {
		MgNet_SetError("Invalid channel");
		return(0);
	}
	return(1);
}

/* Bind the address 'address' to the requested channel on the UDP socket.
   If the channel is -1, then the first unbound channel will be bound with
   the given address as it's primary address.
   If the channel is already bound, this new address will be added to the
   list of valid source addresses for packets arriving on the channel.
   If the channel is not already bound, then the address becomes the primary
   address, to which all outbound packets on the channel are sent.
   This function returns the channel which was bound, or -1 on error.
*/
int MgNet_UDP_Bind(MgNetUDPSocket * sock, int channel, MgNetIPAddress *address)
{
	MgNetUDPSocket::MgNetUDPChannel *binding;

	if ( channel == -1 ) {
		for ( channel=0; channel < MGNET_MAX_UDPCHANNELS; ++channel ) {
			binding = &sock->binding[channel];
			if ( binding->numbound < MGNET_MAX_UDPADDRESSES ) {
				break;
			}
		}
	} else {
		if ( ! ValidChannel(channel) ) {
			return(-1);
		}
		binding = &sock->binding[channel];
	}
	if ( binding->numbound == MGNET_MAX_UDPCHANNELS ) {
		MgNet_SetError("No room for new addresses");
		return(-1);
	}
	binding->address[binding->numbound++] = *address;
	return(channel);
}

/* Unbind all addresses from the given channel */
void MgNet_UDP_Unbind(MgNetUDPSocket * sock, int channel)
{
	if ( (channel >= 0) && (channel < MGNET_MAX_UDPCHANNELS) ) {
		sock->binding[channel].numbound = 0;
	}
}

/* Get the primary IP address of the remote system associated with the
   socket and channel.
   If the channel is not bound, this function returns NULL.
 */
MgNetIPAddress *MgNet_UDP_GetPeerAddress(MgNetUDPSocket * sock, int channel)
{
	MgNetIPAddress *address;

	address = NULL;
	switch (channel) {
		case -1:
			/* Return the actual address of the socket */
			address = &sock->address;
			break;
		default:
			/* Return the address of the bound channel */
			if ( ValidChannel(channel) &&
				(sock->binding[channel].numbound > 0) ) {
				address = &sock->binding[channel].address[0];
			}
			break;
	}
	return(address);
}

/* Send a vector of packets to the the channels specified within the packet.
   If the channel specified in the packet is -1, the packet will be sent to
   the address in the 'src' member of the packet.
   Each packet will be updated with the status of the packet after it has 
   been sent, -1 if the packet send failed.
   This function returns the number of packets sent.
*/
int MgNet_UDP_SendV(MgNetUDPSocket * sock, MgNetUDPPacket **packets, int npackets)
{
	int numsent, i, j;
	MgNetUDPSocket::MgNetUDPChannel *binding;
	int status;
#ifndef MACOS_OPENTRANSPORT
	int sock_len;
	struct sockaddr_in sock_addr;

	/* Set up the variables to send packets */
	sock_len = sizeof(sock_addr);
#endif

	numsent = 0;
	for ( i=0; i<npackets; ++i ) 
	{
		/* if channel is < 0, then use channel specified in sock */
		
		if ( packets[i]->channel < 0 ) 
		{
#ifdef MACOS_OPENTRANSPORT
		TUnitData OTpacket;
		InetAddress address;

			memset(&OTpacket, 0, sizeof(OTpacket));
			OTpacket.addr.buf = (Uint8 *)&address;
			OTpacket.addr.len = (sizeof address);
			OTpacket.udata.buf = packets[i]->data;
			OTpacket.udata.len = packets[i]->len;
			OTInitInetAddress(&address, packets[i]->address.port,packets[i]->address.host);
			
			status = OTSndUData(sock->channel, &OTpacket);
			
			packets[i]->status = status;
			
			if (status == noErr)
			{
				++numsent;
			}
#else
			sock_addr.sin_addr.s_addr = packets[i]->address.host;
			sock_addr.sin_port = packets[i]->address.port;
			sock_addr.sin_family = AF_INET;
			status = sendto(sock->channel, 
#ifdef Win32_Winsock
					(const char *)packets[i]->data,
#else
					packets[i]->data, 
#endif
					packets[i]->len, 0,
					(struct sockaddr *)&sock_addr,sock_len);
			if ( status >= 0 ) {
				packets[i]->status = status;
				++numsent;
			}
#endif /* MACOS_OPENTRANSPORT */
		}
		else 
		{
			/* Send to each of the bound addresses on the channel */
			
			binding = &sock->binding[packets[i]->channel];
			
			for ( j=binding->numbound-1; j>=0; --j ) 
			{
#ifdef MACOS_OPENTRANSPORT
			TUnitData OTpacket;
			InetAddress address;

				memset(&OTpacket, 0, sizeof(OTpacket));
				OTpacket.addr.buf = (Uint8 *)&address;
				OTpacket.addr.len = (sizeof address);
				OTpacket.udata.buf = packets[i]->data;
				OTpacket.udata.len = packets[i]->len;
				OTInitInetAddress(&address, binding->address[j].port,binding->address[j].host);
			                              
				status = OTSndUData(sock->channel, &OTpacket);

				packets[i]->status = status;
				
				if (status == noErr)
				{
					++numsent;
				}

#else
				sock_addr.sin_addr.s_addr = binding->address[j].host;
				sock_addr.sin_port = binding->address[j].port;
				sock_addr.sin_family = AF_INET;
				status = sendto(sock->channel, 
#ifdef Win32_Winsock
						(const char *)packets[i]->data,
#else
						packets[i]->data, 
#endif
						packets[i]->len, 0,
						(struct sockaddr *)&sock_addr,sock_len);
				if ( status >= 0 ) {
					packets[i]->status = status;
					++numsent;
				}
#endif /* MACOS_OPENTRANSPORT */
			}
		}
	}
	
	return(numsent);
}

int MgNet_UDP_Send(MgNetUDPSocket * sock, int channel, MgNetUDPPacket *packet)
{
	/* This is silly, but... */
	packet->channel = channel;
	return(MgNet_UDP_SendV(sock, &packet, 1));
}

/* Returns true if a socket is has data available for reading right now */
static int SocketReady(SOCKET sock)
{
	int retval;
#ifdef MACOS_OPENTRANSPORT
	OTResult status;
	size_t	numBytes;
#else
	struct timeval tv;
	fd_set mask;
#endif

#ifdef MACOS_OPENTRANSPORT
	status = OTCountDataBytes(sock,&numBytes);
	
	if (status == noErr && numBytes > 0)
	{
		retval = 1;
	} 
	
	else 
	{
		retval = 0;
	}
#else
	/* Check the file descriptors for available data */
	do {
		errno = 0;

		/* Set up the mask of file descriptors */
		FD_ZERO(&mask);
		FD_SET(sock, &mask);

		/* Set up the timeout */
		tv.tv_sec = 0;
		tv.tv_usec = 0;

		/* Look! */
		retval = select(sock+1, &mask, NULL, NULL, &tv);
	} while ( errno == EINTR );
#endif /* MACOS_OPENTRANSPORT */

	return(retval == 1);
}

/* Receive a vector of pending packets from the UDP socket.
   The returned packets contain the source address and the channel they arrived
   on.  If they did not arrive on a bound channel, the the channel will be set
   to -1.
   This function returns the number of packets read from the network, or -1
   on error.  This function does not block, so can return 0 packets pending.
*/
 int MgNet_UDP_RecvV(MgNetUDPSocket * sock, MgNetUDPPacket **packets)
{
	int numrecv, i, j;
	MgNetUDPSocket::MgNetUDPChannel *binding;
#ifdef MACOS_OPENTRANSPORT
	TUnitData OTpacket;
	OTFlags flags;
	InetAddress address;
#else
	int sock_len;
	struct sockaddr_in sock_addr;
#endif

	numrecv = 0;
	while ( packets[numrecv] && SocketReady(sock->channel) ) 
	{
	MgNetUDPPacket *packet;

		packet = packets[numrecv];
		
#ifdef MACOS_OPENTRANSPORT
		memset(&OTpacket, 0, sizeof(OTpacket));
		OTpacket.addr.buf = (Uint8 *)&address;
		OTpacket.addr.maxlen = (sizeof address);
		OTpacket.udata.buf = packet->data;
		OTpacket.udata.maxlen = packet->maxlen;
		
		packet->status = OTRcvUData(sock->channel, &OTpacket, &flags);
		
		if (packet->status == noErr)
		{
			packet->len = OTpacket.udata.len;
			packet->address.host = address.fHost;
			packet->address.port = address.fPort;
		}
#else
		sock_len = sizeof(sock_addr);
		packet->status = recvfrom(sock->channel,
#ifdef Win32_Winsock
				(char *)packet->data,
#else
				packet->data, 
#endif
				packet->maxlen, 0,
				(struct sockaddr *)&sock_addr,
#ifdef USE_GUSI_SOCKETS
				(unsigned int *)&sock_len);
#else
# ifdef __linux__
                        (socklen_t *)&sock_len);
# else
						&sock_len);
# endif
#endif
		if ( packet->status >= 0 ) {
			packet->len = packet->status;
			packet->address.host = sock_addr.sin_addr.s_addr;
			packet->address.port = sock_addr.sin_port;
		}
#endif
		if (packet->status >= 0)
		{
			packet->channel = -1;
			
			for (i=(MGNET_MAX_UDPCHANNELS-1); i>=0; --i ) 
			{
				binding = &sock->binding[i];
				
				for ( j=binding->numbound-1; j>=0; --j ) 
				{
					if ( (packet->address.host == binding->address[j].host) &&
					     (packet->address.port == binding->address[j].port) ) 
					{
						packet->channel = i;
						goto foundit; /* break twice */
					}
				}
			}
foundit:
			++numrecv;
		} 
		
		else 
		{
			packet->len = 0;
		}
	}
	
	sock->ready = 0;
	
	return(numrecv);
}

/* Receive a single packet from the UDP socket.
   The returned packet contains the source address and the channel it arrived
   on.  If it did not arrive on a bound channel, the the channel will be set
   to -1.
   This function returns the number of packets read from the network, or -1
   on error.  This function does not block, so can return 0 packets pending.
*/
int MgNet_UDP_Recv(MgNetUDPSocket * sock, MgNetUDPPacket *packet)
{
	MgNetUDPPacket *packets[2];

	/* Receive a packet array of 1 */
	packets[0] = packet;
	packets[1] = NULL;
	return(MgNet_UDP_RecvV(sock, packets));
}

/* Close a UDP network socket */
void MgNet_UDP_Close(MgNetUDPSocket * sock)
{
	if ( sock != NULL ) 
	{
		if ( sock->channel != INVALID_SOCKET ) 
		{
#ifdef MACOS_OPENTRANSPORT
			OTUnbind(sock->channel);
			OTCloseProvider(sock->channel);
#else
			closesocket(sock->channel);
#endif /* MACOS_OPENTRANSPORT */
		}
		
		free(sock);
	}
}
