// MgApp.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#include "MgApp.h"
#include "MgNetLib.h"
#include "internal/MgAppInternal.h"

#include "debug_new.h"

class MgAppThread : public MgThread
{
public:
	MgAppThread(void * data) : MgThread(data) {}
	
protected:
	virtual int Go(void * data)
	{
		MgApp * app = (MgApp *)data;
		
		int ret = app->RunApp();
		Cancel();	// you must cancel!!! otherwise BOOOOOOOOOOOOMMMMMMM
		
		return ret;
	}
};



MgApp::MgApp()
{
	fApp = this;
}

MgApp::~MgApp()
{
	fApp = NULL;
	_dbg_dump_unfreed();
}

int
MgApp::Run(Uint32 flags, const char * title, int argc, char ** argv)
{
	Uint32 sdlFlags = 0;
	int i;
	
	/* check SDL version before doing anything else */
	if (!MgTkCheckVersion())
		return MG_RUN_WRONG_SDL;
	
	fQuit = false;
	fFlags = flags;
	
	if (flags & MG_TIMER)
		sdlFlags |= MG_TIMER;
	if (flags & MG_AUDIO)
		sdlFlags |= MG_AUDIO;
	if (flags & MG_VIDEO)
		sdlFlags |= MG_VIDEO;
	if (flags & MG_CDROM)
		sdlFlags |= MG_CDROM;
	if (flags & MG_JOYSTICK)
		sdlFlags |= MG_JOYSTICK;
	if (flags & MG_NO_PARACHUTE)
		sdlFlags |= MG_NO_PARACHUTE;
	if (flags & MG_EVERYTHING)
		sdlFlags |= MG_EVERYTHING;
	
	if (SDL_Init(sdlFlags) < 0)
		return MG_RUN_SDL;
	
	if (flags & MG_NET)
	{
		if (MgNet_Init() < 0)
		{
			SDL_Quit();
			return MG_RUN_NET;
		}
	}
	if (argc > 1)
	{
		ArgvReceived(argc, argv);
	}
	else
	{
		if (flags & MG_ARGV_RECEIVED_ANYWAY)
			ArgvReceived(argc, argv);
	}

	fVideo = new MgVideoSystem(title);
	fTimer = new MgTimerSystem();
	fAudio = new MgAudioSystem();
	fNet = new MgNetSystem();
	fJoy = new MgJoystickSystem();
	
	// set up filters
	fFilterLock = new MgMutex;
	fNumUserFilters = fNumSystemFilters = 0;
	fUserFilters = fSystemFilters = NULL;

	// we're ready
	ReadyToRun();

	// do this after everything else is done
	fHaveRawEvent = false;

	if (!(flags & MG_NO_EVENTTHREAD))
	{
		fLoopThread = new MgAppThread(this);
		if (fLoopThread->Run() != 0)
		{
			if (flags & MG_NET)
				MgNet_Quit();
			SDL_Quit();
			return MG_RUN_THREAD;
		}
	}	

	if (flags & MG_VIDEO)
	{
		if (!(flags & MG_NO_EVENTTHREAD))
		{
			while (!WasQuit())
			{
				EventLoop();
				SDL_Delay(5);
			}
		}
	}
	
	int ret = 0;
	
	if (flags & MG_NO_EVENTTHREAD)
	{
		ret = RunApp();	// no thread...
	}
	else
	{
		// the loop thread is now destroyed... we can only quit
		// if the user exited in Run();
//		fFilterLock->Lock();
		fLoopThread->Wait();
		ret = fLoopThread->GetReturnValue();
		delete fLoopThread;
//		fFilterLock->Unlock();
	}
	delete fFilterLock;
	// shutdown subsystems
	delete fVideo;
	delete fTimer;
	delete fAudio;
	delete fNet;
	delete fJoy;
	
	if (flags & MG_NET)
		MgNet_Quit();
	
	for (i = 0; i < fNumUserFilters; i++)
		delete fUserFilters[i];
	if (fUserFilters)
		delete fUserFilters;

	for (i = 0; i < fNumSystemFilters; i++)
		delete fSystemFilters[i];
	if (fSystemFilters)
		delete fSystemFilters;

	SDL_Quit();

	return ret;
}

/*!
	This is called after ArgvReceived(). Do some initialization in this method. The default does nothing.
	If you are going to use the video subsystem, set the video mode in this function!
*/
void 
MgApp::ReadyToRun()
{
}

/*!
	If the argc is > 1 in the main() function, this is called. Otherwise it is not unless you
	passed the MG_ARGV_RECEIVED_ANYWAY flag.
*/
void 
MgApp::ArgvReceived(int argc, char ** argv)
{
}

/*!
	This function is called if a SDL_QUIT event was received. Return true to quit, or false to keep running.
	The default returns true. Be warned, if you passed MG_NO_EVENTTHREAD, this function will not be
	called!
*/
bool 
MgApp::QuitRequested()
{
	return true;
}

/*!
	This function is the guts of your app. It should do whatever the app is supposed to do! Reimplement it
	to do just that;
*/
int 
MgApp::RunApp()
{
	return 0;
}

/*!
	Call this function in your Run() function. If it returns true, it means that
	QuitRequested() returned true, so you should quit.
*/
bool 
MgApp::WasQuit()
{
	return fQuit;
}

void
MgApp::Quit()
{
	fQuit = true;
}

/*!
	\retval NULL The video mode has not been set.
	\retval Non-NULL A valid pointer.
*/
MgVideoSystem *
MgApp::Video()
{
	return fVideo;
}

/*!
	\retval NULL The timer system hasn't been started yet.
	\retval Non-NULL A valid pointer.
*/
MgTimerSystem *
MgApp::Timer()
{
	return fTimer;
}

MgAudioSystem *
MgApp::Audio()
{
	return fAudio;
}

MgNetSystem *
MgApp::Net()
{
	return fNet;
}

MgJoystickSystem *
MgApp::Joystick()
{
	return fJoy;
}

/*!
	\retval true Got a new event.
	\retval false No new event returned. The queue is empty.
*/
bool
MgApp::GetRawEvent(SDL_Event & event)
{
	if (fFlags & MG_NO_EVENTTHREAD)
	{
		return (SDL_PollEvent(&event) == 0) ? false : true;
	}
	else
	{
		if (fHaveRawEvent)
		{
			event = fRawEvent;
			fHaveRawEvent = false;
			return true;
		}
	}
	return false;
}

/*!
	\retval true Got an event.
	\retval false An error occured.
*/
bool
MgApp::WaitRawEvent(SDL_Event & event)
{
	if (fFlags & MG_NO_EVENTTHREAD)
	{
		return (SDL_WaitEvent(&event) == 0) ? false : true;
	}
	else
	{
		while (!fHaveRawEvent)
			Delay(2);
		event = fRawEvent;
		fHaveRawEvent = false;
		return true;
	}
}

/*! Be sure you do not call this method at the same time you call MgApp::GetRawEvent()
	or MgApp::WaitRawEvent() because if one of those methods executes first, the event
	they got will be lost to the event filters.
*/
MgApp::EventStatus
MgApp::PumpEvents()
{
	SDL_Event event;
	bool breakLoop = false;
	bool quit = false;	/* flag to see if SDL_QUIT event was received */
	
	if (!fFilterLock->Lock())
		return MGA_LOCK_FAILED;
	
	/* basically, this is the same as MgApp::EventLoop() */
	while (SDL_PollEvent(&event))
	{
		switch (event.type)
		{
			case SDL_QUIT:
				quit = true;
				/* drop through */
			
			default:
			{
				int i = 0;
				for (; i < fNumSystemFilters; i++)
				{
					if (fSystemFilters[i]->mask & SDL_EVENTMASK(event.type))
					{
						switch (event.type)
						{
							case SDL_QUIT:
								fSystemFilters[i]->filter->QuitEvent(MgQuitEvent(event.quit));
								break;
							case SDL_ACTIVEEVENT:
								fSystemFilters[i]->filter->Active(MgActiveEvent(event.active));
								break;
							case SDL_KEYDOWN:
							case SDL_KEYUP:
								fSystemFilters[i]->filter->Keyboard(MgKeyboardEvent(event.key));
								break;
							case SDL_MOUSEMOTION:
								fSystemFilters[i]->filter->MouseMotion(MgMouseMotionEvent(event.motion));
								break;
							case SDL_MOUSEBUTTONUP:
							case SDL_MOUSEBUTTONDOWN:
								fSystemFilters[i]->filter->MouseButton(MgMouseButtonEvent(event.button));
								break;
							case SDL_JOYAXISMOTION:
								fSystemFilters[i]->filter->JoystickAxis(MgJoyAxisEvent(event.jaxis));
								break;
							case SDL_JOYBALLMOTION:
								fSystemFilters[i]->filter->JoystickBall(MgJoyBallEvent(event.jball));
								break;
							case SDL_JOYHATMOTION:
								fSystemFilters[i]->filter->JoystickHat(MgJoyHatEvent(event.jhat));
								break;
							case SDL_JOYBUTTONDOWN:
							case SDL_JOYBUTTONUP:
								fSystemFilters[i]->filter->JoystickButton(MgJoyButtonEvent(event.jbutton));
								break;
							case SDL_VIDEORESIZE:
								fSystemFilters[i]->filter->ResizeEvent(MgResizeEvent(event.resize));
								break;
							case SDL_VIDEOEXPOSE:
								fSystemFilters[i]->filter->ExposeEvent(MgExposeEvent(event.expose));
								break;
							case 24:
							case 25:
							case 26:
							case 27:
							case 28:
							case 29:
							case 30:
							case 31:
								fSystemFilters[i]->filter->UserEvent(MgUserEvent(event.user));
								break;
						}
					}
				}
				for (; i < fNumUserFilters && !breakLoop; i++)
				{
					if (fUserFilters[i]->mask & SDL_EVENTMASK(event.type))
					{
						switch (event.type)
						{
							case SDL_QUIT:
								if (fUserFilters[i]->filter->QuitEvent(MgQuitEvent(event.quit))) breakLoop = true;
								break;
							case SDL_ACTIVEEVENT:
								if (fUserFilters[i]->filter->Active(MgActiveEvent(event.active))) breakLoop = true;
								break;
							case SDL_KEYDOWN:
							case SDL_KEYUP:
								if (fUserFilters[i]->filter->Keyboard(MgKeyboardEvent(event.key))) breakLoop = true;
								break;
							case SDL_MOUSEMOTION:
								if (fUserFilters[i]->filter->MouseMotion(MgMouseMotionEvent(event.motion))) breakLoop = true;
								break;
							case SDL_MOUSEBUTTONUP:
							case SDL_MOUSEBUTTONDOWN:
								if (fUserFilters[i]->filter->MouseButton(MgMouseButtonEvent(event.button))) breakLoop = true;
								break;
							case SDL_JOYAXISMOTION:
								if (fUserFilters[i]->filter->JoystickAxis(MgJoyAxisEvent(event.jaxis))) breakLoop = true;
								break;
							case SDL_JOYBALLMOTION:
								if (fUserFilters[i]->filter->JoystickBall(MgJoyBallEvent(event.jball))) breakLoop = true;
								break;
							case SDL_JOYHATMOTION:
								if (fUserFilters[i]->filter->JoystickHat(MgJoyHatEvent(event.jhat))) breakLoop = true;
								break;
							case SDL_JOYBUTTONDOWN:
							case SDL_JOYBUTTONUP:
								if (fUserFilters[i]->filter->JoystickButton(MgJoyButtonEvent(event.jbutton))) breakLoop = true;
								break;
							case SDL_VIDEORESIZE:
								if (fUserFilters[i]->filter->ResizeEvent(MgResizeEvent(event.resize))) breakLoop = true;
								break;
							case SDL_VIDEOEXPOSE:
								if (fUserFilters[i]->filter->ExposeEvent(MgExposeEvent(event.expose))) breakLoop = true;
								break;
							case 24:
							case 25:
							case 26:
							case 27:
							case 28:
							case 29:
							case 30:
							case 31:
								if (fUserFilters[i]->filter->UserEvent(MgUserEvent(event.user))) breakLoop = true;
								break;
						}
					}
				}
				break;
			}
		}
	}
	
	fFilterLock->Unlock();
	
	return quit ? MGA_QUIT : MGA_OK;
}

int 
MgApp::EventLoop()
{
	SDL_Event event;
	bool breakLoop = false;
	
	if (!fFilterLock->Lock())
		return 0;

	while (SDL_PollEvent(&event))
	{
		fHaveRawEvent = true;
		fRawEvent = event;
		switch (event.type)
		{
			case SDL_QUIT:
				fQuit = QuitRequested();
				/* drop through */
				
			default:
			{
				int i = 0;
				for (; i < fNumSystemFilters; i++)
				{
					if (fSystemFilters[i]->mask & SDL_EVENTMASK(event.type))
					{
						switch (event.type)
						{
							case SDL_QUIT:
								fSystemFilters[i]->filter->QuitEvent(MgQuitEvent(event.quit));
								break;
							case SDL_ACTIVEEVENT:
								fSystemFilters[i]->filter->Active(MgActiveEvent(event.active));
								break;
							case SDL_KEYDOWN:
							case SDL_KEYUP:
								fSystemFilters[i]->filter->Keyboard(MgKeyboardEvent(event.key));
								break;
							case SDL_MOUSEMOTION:
								fSystemFilters[i]->filter->MouseMotion(MgMouseMotionEvent(event.motion));
								break;
							case SDL_MOUSEBUTTONUP:
							case SDL_MOUSEBUTTONDOWN:
								fSystemFilters[i]->filter->MouseButton(MgMouseButtonEvent(event.button));
								break;
							case SDL_JOYAXISMOTION:
								fSystemFilters[i]->filter->JoystickAxis(MgJoyAxisEvent(event.jaxis));
								break;
							case SDL_JOYBALLMOTION:
								fSystemFilters[i]->filter->JoystickBall(MgJoyBallEvent(event.jball));
								break;
							case SDL_JOYHATMOTION:
								fSystemFilters[i]->filter->JoystickHat(MgJoyHatEvent(event.jhat));
								break;
							case SDL_JOYBUTTONDOWN:
							case SDL_JOYBUTTONUP:
								fSystemFilters[i]->filter->JoystickButton(MgJoyButtonEvent(event.jbutton));
								break;
							case SDL_VIDEORESIZE:
								fSystemFilters[i]->filter->ResizeEvent(MgResizeEvent(event.resize));
								break;
							case SDL_VIDEOEXPOSE:
								fSystemFilters[i]->filter->ExposeEvent(MgExposeEvent(event.expose));
								break;
							case 24:
							case 25:
							case 26:
							case 27:
							case 28:
							case 29:
							case 30:
							case 31:
								fSystemFilters[i]->filter->UserEvent(MgUserEvent(event.user));
								break;
						}
					}
				}
				for (; i < fNumUserFilters && !breakLoop; i++)
				{
					if (fUserFilters[i]->mask & SDL_EVENTMASK(event.type))
					{
						switch (event.type)
						{
							case SDL_QUIT:
								if (fUserFilters[i]->filter->QuitEvent(MgQuitEvent(event.quit))) breakLoop = true;
								break;
							case SDL_ACTIVEEVENT:
								if (fUserFilters[i]->filter->Active(MgActiveEvent(event.active))) breakLoop = true;
								break;
							case SDL_KEYDOWN:
							case SDL_KEYUP:
								if (fUserFilters[i]->filter->Keyboard(MgKeyboardEvent(event.key))) breakLoop = true;
								break;
							case SDL_MOUSEMOTION:
								if (fUserFilters[i]->filter->MouseMotion(MgMouseMotionEvent(event.motion))) breakLoop = true;
								break;
							case SDL_MOUSEBUTTONUP:
							case SDL_MOUSEBUTTONDOWN:
								if (fUserFilters[i]->filter->MouseButton(MgMouseButtonEvent(event.button))) breakLoop = true;
								break;
							case SDL_JOYAXISMOTION:
								if (fUserFilters[i]->filter->JoystickAxis(MgJoyAxisEvent(event.jaxis))) breakLoop = true;
								break;
							case SDL_JOYBALLMOTION:
								if (fUserFilters[i]->filter->JoystickBall(MgJoyBallEvent(event.jball))) breakLoop = true;
								break;
							case SDL_JOYHATMOTION:
								if (fUserFilters[i]->filter->JoystickHat(MgJoyHatEvent(event.jhat))) breakLoop = true;
								break;
							case SDL_JOYBUTTONDOWN:
							case SDL_JOYBUTTONUP:
								if (fUserFilters[i]->filter->JoystickButton(MgJoyButtonEvent(event.jbutton))) breakLoop = true;
								break;
							case SDL_VIDEORESIZE:
								if (fUserFilters[i]->filter->ResizeEvent(MgResizeEvent(event.resize))) breakLoop = true;
								break;
							case SDL_VIDEOEXPOSE:
								if (fUserFilters[i]->filter->ExposeEvent(MgExposeEvent(event.expose))) breakLoop = true;
								break;
							case 24:
							case 25:
							case 26:
							case 27:
							case 28:
							case 29:
							case 30:
							case 31:
								if (fUserFilters[i]->filter->UserEvent(MgUserEvent(event.user))) breakLoop = true;
								break;
						}
					}
				}
				break;
			}
		}
	}
	fFilterLock->Unlock();
	return 0;
}

/*!	\param filter The filter.
	\param events A mask of the events defined in MgEventFilter.h. You must implement their respective methods in MgEventFilter to catch that event.
*/
bool 
MgApp::AddFilter(MgEventFilter * filter, Uint32 events, int type)
{
	bool ret = false;
	int i;
	
	switch (type)
	{
		case MG_FILTER_USER:
		{
			MgAppFilterHelper ** filters = new MgAppFilterHelper *[fNumUserFilters + 1];
		
			if (fFilterLock->Lock())
			{	
				for (i = 0; i < fNumUserFilters; i++)
					filters[i] = fUserFilters[i];
				filters[i] = new MgAppFilterHelper;
				filters[i]->filter = filter;
				filters[i]->mask = events;
//				for (i = 0; i < fNumUserFilters; i++)
//					delete fUserFilters[i];
				delete fUserFilters;
				fUserFilters = filters;
				fNumUserFilters++;
				ret = true;
				fFilterLock->Unlock();
			}
			break;
		}
		
		case MG_FILTER_SYSTEM:
		{
			MgAppFilterHelper ** filters = new MgAppFilterHelper *[fNumSystemFilters + 1];
		
			if (fFilterLock->Lock())
			{	
				for (i = 0; i < fNumSystemFilters; i++)
					filters[i] = fSystemFilters[i];
				filters[i] = new MgAppFilterHelper;
				filters[i]->filter = filter;
				filters[i]->mask = events;
//				for (i = 0; i < fNumSystemFilters; i++)
//					delete fSystemFilters[i];
				delete fSystemFilters;
				fSystemFilters = filters;
				fNumSystemFilters++;
				ret = true;
				fFilterLock->Unlock();
			}
			break;
		}
	}
	return ret;
}

bool 
MgApp::RemoveFilter(MgEventFilter * filter, int type)
{
	MgAppFilterHelper ** filters;
	int i = 0;
	bool ret = false;
	if (!fFilterLock->Lock())
		return false;

	switch (type)
	{
		case MG_FILTER_USER:
		{		
			for (; i < fNumUserFilters; i++)
			{
				if (fUserFilters[i]->filter == filter)
					break;
			}
			if (i == fNumUserFilters)
				return false;
			
			delete fUserFilters[i];
			
			filters = new MgAppFilterHelper *[fNumUserFilters - 1];
			for (int j = 0; j < fNumUserFilters; j++)
			{
				if (j < i)
					filters[j] = fUserFilters[j];
				else if (j > i)
					filters[j - 1] = fUserFilters[j];
			}
			delete fUserFilters;
			fUserFilters = filters;
			fNumUserFilters--;
			fFilterLock->Unlock();
			ret = true;
			break;
		}

		case MG_FILTER_SYSTEM:
		{
			for (; i < fNumSystemFilters; i++)
			{
				if (fSystemFilters[i]->filter == filter)
					break;
			}
			if (i == fNumSystemFilters)
				return false;
			
			delete fSystemFilters[i];
			
			filters = new MgAppFilterHelper *[fNumSystemFilters - 1];
			for (int j = 0; j < fNumSystemFilters; j++)
			{
				if (j < i)
					filters[j] = fSystemFilters[j];
				else if (j > i)
					filters[j - 1] = fSystemFilters[j];
			}
			delete fSystemFilters;
			fSystemFilters = filters;
			fNumSystemFilters--;
			fFilterLock->Unlock();
			ret = true;
			break;
		}
	}
	return ret;
}

/*! This method acquires the lock using the filter mutex, which
	effectively stops the filter thread from running. Be sure
	to resume the filter when you want it to run again!
*/
bool
MgApp::PauseFilter()
{
	return fFilterLock->Lock();
}

void
MgApp::ResumeFilter()
{
	fFilterLock->Unlock();
}

Uint32
MgApp::GetTicks()
{
	return SDL_GetTicks();
}

void
MgApp::Delay(Uint32 ms)
{
	SDL_Delay(ms);
}

MgApp *
MgApp::App()
{
	return fApp;
}


MgVideoSystem * MgApp::fVideo = NULL;
MgTimerSystem * MgApp::fTimer = NULL;
MgAudioSystem * MgApp::fAudio = NULL;
MgNetSystem * MgApp::fNet = NULL;
MgJoystickSystem * MgApp::fJoy = NULL;

MgApp * MgApp::fApp = NULL;


/* version checking */

/* The minimum SDL version required for libMgTk */
#define MG_SDL_MAJOR_VERSION 1
#define MG_SDL_MINOR_VERSION 2
#define MG_SDL_PATCH_LEVEL   2

bool
MgTkCheckVersion()
{
	const SDL_version * ver = SDL_Linked_Version();
	
	printf("SDL Version: %d.%d.%d\n", ver->major, ver->minor, ver->patch);
	
	if (ver->major >= MG_SDL_MAJOR_VERSION)
	{
		if (ver->minor >= MG_SDL_MINOR_VERSION)
		{
			if (ver->patch >= MG_SDL_PATCH_LEVEL)
			{
				return true;
			}
		}
	}
	
	return false;
}
