// MgInit.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#include "MgInit.h"
#include "MgAutoDelete.h"
#include "debug_new.h"
#include "MgNetLib.h"
#include "internal/MgAppInternal.h"

/* default app */
class MgDefaultApp : public MgApp
{
public:
	MgDefaultApp();
	virtual ~MgDefaultApp();
	
	virtual int Run(Uint32 flags, const char * title, int argc, char ** argv);
	
	//virtual bool AddFilter(MgEventFilter * filter, Uint32 events, int type = MG_FILTER_USER);
	//virtual bool RemoveFilter(MgEventFilter * filter, int type = MG_FILTER_USER);
	virtual bool PauseFilter();
	virtual void ResumeFilter();
	
	virtual bool GetRawEvent(SDL_Event & event);
	virtual bool WaitRawEvent(SDL_Event & event);

private:
};

/* this is a dummy class with only GetRawEvent() and WaitRawEvent() doing anything */
MgDefaultApp::MgDefaultApp()
	: MgApp()
{
	/* init the event filter vars */
	fFilterLock = new MgMutex;
	fNumUserFilters = fNumSystemFilters = 0;
	fUserFilters = fSystemFilters = NULL;
}

MgDefaultApp::~MgDefaultApp()
{
	int i;
	
	delete fFilterLock;
	for (i = 0; i < fNumUserFilters; i++)
		delete fUserFilters[i];
	delete fUserFilters;
	for (i = 0; i < fNumSystemFilters; i++)
		delete fSystemFilters[i];
	delete fSystemFilters;
}

int 
MgDefaultApp::Run(Uint32 flags, const char * title, int argc, char ** argv)
{
	return 0;
}

//bool 
//MgDefaultApp::AddFilter(MgEventFilter * filter, Uint32 events, int type)
//{
//	return true;
//}
//
//bool 
//MgDefaultApp::RemoveFilter(MgEventFilter * filter, int type)
//{
//	return true;
//}

bool 
MgDefaultApp::PauseFilter()
{
	return true;
}

void 
MgDefaultApp::ResumeFilter()
{
}

bool 
MgDefaultApp::GetRawEvent(SDL_Event &  event)
{
	return (SDL_PollEvent(&event) == 0) ? false : true;
}

bool 
MgDefaultApp::WaitRawEvent(SDL_Event & event)
{
	return (SDL_WaitEvent(&event) == 0) ? false : true;
}

/* end default app */


/* static globals */
static MgSystems * mgtkSystems = NULL;
static MgDefaultApp * mgtkApp = NULL;

typedef MgDelete<MgSubSystem *> MgDeleteSys;

#define FAIL() delete mgtkSystems; mgtkSystems = NULL; SDL_Quit(); return NULL

const MgSystems * const
MgTkInit(Uint32 flags)
{
	if (mgtkSystems)	/* have we been initialized yet? */
		return mgtkSystems;		/* yes */
	
	/* see if we have an MgApp running */
	if (MgApp::App())
		return MG_INIT_RUNNING; /* we are running already */
	
	/* check SDL version */
	if (!MgTkCheckVersion())
	{
		return MG_INIT_WRONG_SDL;
	}

	/* start SDL */
	if (SDL_Init(flags) < 0)
		return NULL;

	if (flags & MG_NET)
	{
		if (MgNet_Init() != 0)
		{
			SDL_Quit();
			return NULL;
		}
	}
	
	/* create app */
	mgtkApp = new MgDefaultApp;
	if (!mgtkApp)
	{
		FAIL();
	}	
	/* app deleter */
	MgDelete<MgDefaultApp *> dApp(mgtkApp);
	
	/* create systems object */
	mgtkSystems = new MgSystems;
	if (!mgtkSystems)
	{
		FAIL();
	}
	
	mgtkSystems->fTimer = new MgTimerSystem;
	if (!mgtkSystems->fTimer)
	{
		FAIL();
	}
	MgDeleteSys dTimer(mgtkSystems->fTimer);
	
	mgtkSystems->fAudio = new MgAudioSystem;
	if (!mgtkSystems->fAudio)
	{
		FAIL();
	}
	MgDeleteSys dAudio(mgtkSystems->fAudio);
	
	mgtkSystems->fVideo = new MgVideoSystem("Default");
	if (!mgtkSystems->fVideo)
	{
		FAIL();
	}
	MgDeleteSys dVideo(mgtkSystems->fVideo);
	
	mgtkSystems->fNet = new MgNetSystem;
	if (!mgtkSystems->fNet)
	{
		FAIL();
	}
	MgDeleteSys dNet(mgtkSystems->fNet);
	
	mgtkSystems->fJoy = new MgJoystickSystem;
	if (!mgtkSystems->fJoy)
	{
		FAIL();
	}
	
	dVideo.DoNotDelete();
	dAudio.DoNotDelete();
	dTimer.DoNotDelete();
	dNet.DoNotDelete();
	dApp.DoNotDelete();
		
	mgtkSystems->fApp = mgtkApp;
	
	return mgtkSystems;
}

void
MgTkQuit()
{
	/* only quit if we were started using MgTkInit() */
	if (mgtkSystems)
	{
		if (mgtkSystems->fNet->IsRunning())
			mgtkSystems->fNet->Quit();

		delete mgtkSystems->fAudio;
		delete mgtkSystems->fVideo;
		delete mgtkSystems->fNet;
		delete mgtkSystems->fTimer;
		delete mgtkSystems->fJoy;
		delete mgtkSystems;
		mgtkSystems = NULL;
		delete mgtkApp;
		mgtkApp = NULL;
		
		SDL_Quit();
	}
}
