// MgAudioSystem.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#include <SDL.h>

#include "MgAudioSystem.h"
#include "MgApp.h"

#include "debug_new.h"

bool 
MgAudioSystem::IsRunning() const
{
	return SDL_WasInit(MG_AUDIO) == MG_AUDIO ? true : false;
}

bool 
MgAudioSystem::Init()
{
	bool ret = IsRunning();
	if (!ret)
		ret = SDL_InitSubSystem(MG_AUDIO) == 0 ? true : false;
	return ret;
}

/*! If you have an MgMixer running, delete it first so that
	it will close the audio device! Then call Quit().
*/
void 
MgAudioSystem::Quit()
{
	if (IsRunning())
		SDL_QuitSubSystem(MG_AUDIO);
}

/*! \param desired The desirec format.
	\param obtained The obtained format.
	
	\return True if successful, otherwise false.
*/
bool 
MgAudioSystem::OpenAudio(SDL_AudioSpec & desired, SDL_AudioSpec & obtained)
{
	return SDL_OpenAudio(&desired, &obtained) == 0 ? true : false;
}

MgAudioStatus 
MgAudioSystem::GetAudioStatus() const
{
	return (MgAudioStatus)SDL_GetAudioStatus();
}

/*! \param pause If true, audio will be paused. If false, playback will be started. */
void 
MgAudioSystem::PauseAudio(bool pause)
{
	SDL_PauseAudio((pause ? 1 : 0));
}

/*!
	\param file The WAV file.
	\param spec The spec.
	\param audioBuf The loaded WAV data will stored here.
	\param audioLen The length of 'audioBuf'.
	
	\return True if successful, false otherwise.
*/
bool 
MgAudioSystem::LoadWAV(const char * file, SDL_AudioSpec & spec, Uint8 ** audioBuf, Uint32 & audioLen)
{
	return SDL_LoadWAV(file, &spec, audioBuf, &audioLen) == NULL ? false : true;
}

/*! \param audioBuf The buffer to free. */
void 
MgAudioSystem::FreeWAV(Uint8 * audioBuf)
{
	SDL_FreeWAV(audioBuf);
}

/*!
	\param cvt This will be initialized with the correct format.
	\param srcFormat/dstFormat Source/dest format.
	\param srcChannels/dstChannels Source/dest channels.
	\param srcRate/dstRate Source/dest rate.
	
	\return True if successfull, false otherwise.
*/
bool 
MgAudioSystem::BuildAudioCVT(SDL_AudioCVT & cvt, Uint16 srcFormat, Uint8 srcChannels, int srcRate, Uint16 dstFormat, Uint8 dstChannels, int dstRate)
{
	// SDL_BuldAudioCVT returns 1 if success, otherwise it returns -1.
	return SDL_BuildAudioCVT(&cvt, srcFormat, srcChannels, srcRate, dstFormat, dstChannels, dstRate) == 1 ? true : false;
}

/*! \param cvt The format to conver to. */
int 
MgAudioSystem::ConvertAudio(SDL_AudioCVT & cvt)
{
	return SDL_ConvertAudio(&cvt);
}

/*!
	\param dst The stream to mix to.
	\param src The source stream.
	\param len The length.
	\param volume The volume.
*/
void 
MgAudioSystem::MixAudio(Uint8 * dst, const Uint8 * src, Uint32 len, int volume)
{
#ifdef WIN32
	SDL_MixAudio(dst, (Uint8 *)src, len, volume);
#else
	SDL_MixAudio(dst, src, len, volume);
#endif
}

void 
MgAudioSystem::CloseAudio()
{
	SDL_CloseAudio();
}

void 
MgAudioSystem::LockAudio()
{
	SDL_LockAudio();
}

void 
MgAudioSystem::UnlockAudio()
{
	SDL_UnlockAudio();
}

