// MgMixer.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#include <stdio.h>

#include "MgMixer.h"
#include "MgApp.h"
#include "internal/MgChannel.h"

#include "debug_new.h"

MgChannel ** MgMixer::fMixCh = NULL;
SDL_AudioSpec MgMixer::fMixerSpec;
MgMutex * MgMixer::fMixerLock = NULL;
Uint32 MgMixer::fOpened = 0;
int MgMixer::fChannels = 8;
MgAudioSystem * MgMixer::fSystem = NULL;

/*!
	\param freq Frequency.
	\param format Audio format.
	\param channels Number of channels.
	\param samples Samples per second.
*/
MgMixer::MgMixer(int freq, Uint16 format, int channels, Uint16 samples)
{
	if (fOpened)
	{
		fOpened++;
		return;
	}
	
	SDL_AudioSpec des;
	des.freq = freq;
	des.format = format;
	des.channels = channels;
	des.samples = samples;
	des.callback = MixerCallback;
	des.userdata = NULL;
	
	fSystem = MgApp::Audio();
	if (!fSystem->OpenAudio(des, fMixerSpec))
		return;
		
	fMixerLock = new MgMutex();
	if (!fMixerLock)
	{
		fSystem->CloseAudio();
		return;
	}
	
	fMixCh = new MgChannel *[fChannels];
	if (!fMixCh)
	{
		fSystem->CloseAudio();
		delete fMixerLock;
		fMixerLock = NULL;
		return;
	}
	
	for (int i = 0; i < fChannels; i++)
	{
		fMixCh[i] = new MgChannel(true, MG_MIX_MAXVOLUME, 0, NULL);
		if (fMixCh[i] == NULL)
		{
			for (int j = i - 1; j >= 0; j--)
				delete fMixCh[j];
			delete fMixCh;
			fMixCh = NULL;
			fSystem->CloseAudio();
			delete fMixerLock;
			fMixerLock = NULL;
		}
	}
	fOpened = 1;
	
	// start playing
	fSystem->PauseAudio(false);
}


MgMixer::~MgMixer()
{
	if (fOpened > 0)
		fOpened--;
	if (fOpened)
		return;	// still at least one MgMixer existing
	
	Stop(-1);
	fSystem->CloseAudio();
	for (int i = 0; i < fChannels; i++)
		delete fMixCh[i];
	delete fMixCh;
	fMixCh = NULL;
	delete fMixerLock;
	fMixerLock = NULL;
	fSystem = NULL;
}

void 
MgMixer::GetSpec(int & freq, Uint16 & format, int & channels) const
{
	freq = fMixerSpec.freq;
	format = fMixerSpec.format;
	channels = fMixerSpec.channels;
}

const SDL_AudioSpec *
MgMixer::GetSpec()
{
	return &fMixerSpec;
}

/*! If channel is -1, change the volume of all channels. */
void 
MgMixer::Volume(int channel, int volume)
{
	if (channel == -1)
	{
		for (int i = 0; i < fChannels; i++)
			Volume(i, volume);
	}
	else
	{
		if (volume < 0)
			volume = 0;
		if (volume > MG_MIX_MAXVOLUME)
			volume = MG_MIX_MAXVOLUME;
		if (fMixerLock->Lock())
		{
			fMixCh[channel]->fVolume = volume;
			fMixerLock->Unlock();
		}
	}
}

/*! If channel is -1, stop all channels. */
void 
MgMixer::Stop(int channel)
{
	if (channel == -1)
	{
		for (int i = 0; i < fChannels; i++)
			Stop(i);
	}
	else
	{
		if (fMixerLock->Lock())
		{
			fMixCh[channel]->fPaused = true;
			fMixCh[channel]->fSoundLeft = 0;
			fMixCh[channel]->fLoops = 0;
			fMixCh[channel]->fVolume = MG_MIX_MAXVOLUME;
			fMixCh[channel]->fSound = NULL;
			fMixerLock->Unlock();
		}
	}
}

void 
MgMixer::Pause(int channel)
{
	if (channel == -1)
	{
		for (int i = 0; i < fChannels; i++)
			Pause(i);
	}
	else
	{
		if (fMixerLock->Lock())
		{
			fMixCh[channel]->fPaused = true;
			fMixerLock->Unlock();
		}
	}
}

void 
MgMixer::Unpause(int channel)
{
	if (channel == -1)
	{
		for (int i = 0; i < fChannels; i++)
			Unpause(i);
	}
	else
	{
		if (fMixerLock->Lock())
		{
			if (fMixCh[channel]->fSoundLeft > 0)
				fMixCh[channel]->fPaused = false;
			fMixerLock->Unlock();
		}
	}
}

int 
MgMixer::Play(int channel, MgSound * sound, int loops)
{
	if (sound == NULL)
		return -1;
	
	int i = -1;
	
	if (fMixerLock->Lock())
	{
		if (channel == -1)
		{
			for (i = 0; i < fChannels; i++)
			{
				if (fMixCh[i]->fSoundLeft == 0)
					break;
			}
			if (i == fChannels)
				return -1;
		}
		else
			i = channel;
		fMixCh[i]->fPaused = false;
		fMixCh[i]->fSoundLeft = sound->SoundLength();
		fMixCh[i]->fLoops = loops;
		fMixCh[i]->fVolume = MG_MIX_MAXVOLUME;
		fMixCh[i]->fSound = sound;
		fMixerLock->Unlock();
	}
	return i;
}

int 
MgMixer::SetChannels(int num)
{
	if (num < 0 || num == fChannels)
		return fChannels;
	
	int i;
	
	if (fMixerLock->Lock())
	{
		if (num < fChannels)
		{
			for (i = num; i < fChannels; i++)
			{
				fMixCh[i]->fSound = NULL;
				delete fMixCh[i];
			}
		}
		else
		{
			MgChannel ** newch = new MgChannel *[num];
			for (i = 0; i < fChannels; i++)
				newch[i] = fMixCh[i];
			for (i = fChannels; i < num; i++)
				newch[i] = new MgChannel(true, MG_MIX_MAXVOLUME, 0, NULL);
			delete fMixCh;
			fMixCh = newch;
		}
		fChannels = num;
		fMixerLock->Unlock();
	}
	return fChannels;
}

void 
MgMixer::MixerCallback(void * , Uint8 * stream, int len)
{
//	Uint8 * waveptr;
	int loopLen;
	Uint8 * loopStream;
	
	if (fMixerLock->Lock())
	{
		for (int i = 0; i < fChannels; i++)
		{
			if (!fMixCh[i]->fPaused)
			{
				if (fMixCh[i]->fSoundLeft > 0)
				{
//					waveptr = fMixCh[i]->fSound->Sound() + fMixCh[i]->fSound->SoundPos();
					if (fMixCh[i]->fLoops > 0)
					{
						loopLen = len;
						loopStream = stream;
						while (fMixCh[i]->fSoundLeft <= loopLen)
						{
							if (fMixCh[i]->fLoops == 0)
								break;
							fMixCh[i]->fSound->Mix(loopStream, fMixCh[i]->fSoundLeft, fMixCh[i]->fVolume);
							loopStream += fMixCh[i]->fSoundLeft;
							loopLen -= fMixCh[i]->fSoundLeft;
							// reset amount left
							fMixCh[i]->fSoundLeft = fMixCh[i]->fSound->SoundLength();
							fMixCh[i]->fSound->Rewind();
							fMixCh[i]->fLoops--;
						}
						if (fMixCh[i]->fLoops > 0 && loopLen > 0)
						{
							fMixCh[i]->fSound->Mix(loopStream, loopLen, fMixCh[i]->fVolume);
							fMixCh[i]->fSoundLeft -= loopLen;
						}
					}
					else
					{
						if (fMixCh[i]->fSoundLeft < len)
						{
							fMixCh[i]->fSound->Mix(stream, fMixCh[i]->fSoundLeft, fMixCh[i]->fVolume);
							fMixCh[i]->fPaused = true;
							fMixCh[i]->fSoundLeft = 0;
							fMixCh[i]->fVolume = MG_MIX_MAXVOLUME;
							fMixCh[i]->fSound->Rewind();
							fMixCh[i]->fSound = NULL;
						}
						else
						{
							fMixCh[i]->fSound->Mix(stream, len, fMixCh[i]->fVolume);
							fMixCh[i]->fSoundLeft -= len;
						}
					}
				}
			}
		}
		fMixerLock->Unlock();
	}
}

bool 
MgMixer::InitCheck() const
{
	if (!fSystem || !fMixCh || !fMixerLock)
		return false;
	return true;
}

