// MgMixer.h
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#ifndef MGMIXER_H
#define MGMIXER_H

#include "MgMutex.h"
#include "MgSound.h"
#include "MgAudioSystem.h"

#include "MgBeginCode.h"

/*! \file MgMixer.h
	Contains the mixer class.
*/

class MgChannel;

/*!
	You must create an MgMixer before loading anyway sounds. This
	is so because certain formats may need to call the MgMixer::GetSpec()
	method to convert the audio internally.
	
	You can create more than one mixer, but those mixers will have the
	same audio spec as the first one. If you destroy all the running mixers
	and then create a new one with a different format, you need to free
	any sounds you have allocated and reload them.
	
	The mixer starts with 8 channels when created. I do not recommend have
	more than 8.
*/
class MGDLL MgMixer
{
public:
	//! Constructor.
	MgMixer(int freq = 22050, Uint16 format = AUDIO_S16, int channels = 2,
			Uint16 samples = 2048);
	//! Destructor.
	~MgMixer();
	
	//! The the current audio spec.
	void GetSpec(int & freq, Uint16 & format, int & channels) const;
	//! Static function to get the audio spec.
	static const SDL_AudioSpec * GetSpec();
	
	//! Change the volume of a channel.
	void Volume(int channel, int volume);
	//! Stop a channel.
	void Stop(int channel);
	//! Pause a channel.
	void Pause(int channel);
	//! Unpause a channel.
	void Unpause(int channel);
	
	//! Play a sound.
	int Play(int channel, MgSound * sound, int loops = 0);
	//! Set the number of channels.
	int SetChannels(int num);
	
	//! Call this before using the mixer!
	bool InitCheck() const;
	
private:
	static MgChannel ** fMixCh;
	static SDL_AudioSpec fMixerSpec;
	static MgMutex * fMixerLock;
	static Uint32 fOpened;
	static int fChannels;
	static MgAudioSystem * fSystem;
	
	static void MixerCallback(void *, Uint8 * stream, int len);
};

#include "MgEndCode.h"

#endif

