// MgCDROM.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/
#include "MgCDROM.h"

/*! \param drive The drive to open. If '-1', then don't open any drive. */
MgCDROM::MgCDROM(int drive)
{
	fDrive = drive;
	if (fDrive < 0)
	{
		fCD = NULL;
	}
	else
	{
		if (!Open(fDrive))
			fCD = NULL;
	}
}

/*! Closes the drive. */
MgCDROM::~MgCDROM()
{
	if (fCD != NULL)
		Close();
}

int 
MgCDROM::NumDrives()
{
	return SDL_CDNumDrives();
}

/*! \param drive The drive index. */
const char *
MgCDROM::Name(int drive)
{
	return SDL_CDName(drive);
}

const char *
MgCDROM::Name()
{
	return fName;
}

/*! You can reopen a currently open drive, or open a different drive. You must
	use this method if you didn't open a drive in the constructor.
	
	\param drive The drive index.
	
	\retval true Success.
	\retval false Failure.
*/
bool 
MgCDROM::Reopen(int drive)
{
	if (fCD != NULL)
		Close();
	fDrive = drive;
	if (!Open(fDrive))
	{
		fCD = NULL;
		return false;
	}
	
	return true;
}

int 
MgCDROM::NumTracks()
{
	return fCD->numtracks;
}

int 
MgCDROM::CurrentTrack()
{
	/* update the status */
	GetStatus();
	return fCD->cur_track;
}

/*! You must call this on a regular basis to update the CD structure with
	the most current values.
*/
MgCDROM::Status 
MgCDROM::GetStatus()
{
	return (MgCDROM::Status)SDL_CDStatus(fCD);
}

bool 
MgCDROM::CDInDrive()
{
	return (GetStatus() > 0) ? true : false;
}

/*!
	\param startTrack The starting track.
	\param startFrame The starting frame.
	\param ntracks Number of tracks to play.
	\param nframes Number of frames to play.
	
	\retval true Playing has started.
	\retval false An error occured.
*/
bool 
MgCDROM::PlayTracks(int startTrack, int startFrame, int ntracks, int nframes)
{
	return (SDL_CDPlayTracks(fCD, startTrack, startFrame, ntracks, nframes) == 0) ?
			true : false;
}

/*!
	\param startTrack Starting track.
	\param lastTrack Last track to play.
	
	\retval true Playing.
	\retval false Error!
*/
bool 
MgCDROM::PlayTracks(int startTrack, int lastTrack)
{
	return PlayTracks(startTrack, 0, (lastTrack - startTrack + 1), 0);
}

/*!
	\param start Starring frame.
	\param lenght Length, in frames, to play.
	
	\retval true Playing has started.
	\retval false Error!
*/
bool 
MgCDROM::Play(int start, int length)
{
	return (SDL_CDPlay(fCD, start, length) == 0) ? true : false;
}

/*!
	\param track The track to play.
	
	\retval true Playing.
	\retval false Error!
*/
bool 
MgCDROM::Play(int track)
{
	return Play((int)(Track(track).offset), (int)(Track(track).length));
}

bool 
MgCDROM::Pause()
{
	return (SDL_CDPause(fCD) == 0) ? true : false;
}

bool 
MgCDROM::Resume()
{
	return (SDL_CDResume(fCD) == 0) ? true : false;
}

bool 
MgCDROM::Stop()
{
	return (SDL_CDStop(fCD) == 0) ? true : false;
}

bool 
MgCDROM::Eject()
{
	return (SDL_CDEject(fCD) == 0) ? true : false;
}

void 
MgCDROM::Close()
{
	SDL_CDClose(fCD);
}

bool 
MgCDROM::Open(int drive)
{
	fCD = SDL_CDOpen(drive);
	if (!fCD)
	{
		fName = NULL;
		return false;
	}
	fName = SDL_CDName(drive);
	return true;
}

/*!
	\param track The track to retrieve the structure for.
*/
MgCDTrack 
MgCDROM::Track(int track)
{
	return fCD->track[track];
}

bool 
MgCDROM::InitCheck() const
{
	return (fCD != NULL) ? true : false;
}

