// MgJoystick.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/
#include "MgJoystick.h"
#include "MgApp.h"

#include <SDL_joystick.h>


int 
MgJoystickBase::Amount()
{
	return SDL_NumJoysticks();
}

const char *
MgJoystickBase::Name(int joystick)
{
	return SDL_JoystickName(joystick);
}

void 
MgJoystickBase::Enable()
{
	SDL_JoystickEventState(SDL_ENABLE);
}

void 
MgJoystickBase::Disable()
{
	SDL_JoystickEventState(SDL_DISABLE);
}

int 
MgJoystickBase::NumAxes()
{
	return SDL_JoystickNumAxes(fJoy);
}

int 
MgJoystickBase::NumBalls()
{
	return SDL_JoystickNumBalls(fJoy);
}

int 
MgJoystickBase::NumHats()
{
	return SDL_JoystickNumHats(fJoy);
}

int 
MgJoystickBase::NumButtons()
{
	return SDL_JoystickNumButtons(fJoy);
}


MgJoystickBase::MgJoystickBase(int joystick)
{
	fError = OpenFailed;
	if (joystick < 0)	/* open the first unopened joystick? */
	{
		for (joystick = 0; joystick < SDL_NumJoysticks(); joystick++)
		{
			if (!SDL_JoystickOpened(joystick))
				break;
		}
	}
	else
	{
		if (SDL_JoystickOpened(joystick) == 1)	/* returns 1 if joystick was opened */
		{
			fError = AlreadyOpen;
			return;
		}
	}
	fJoy = SDL_JoystickOpen(joystick);
	if (fJoy)
		fError = NoError;
}


MgJoystickBase::~MgJoystickBase()
{
	/* close the joystick if it was ever opened */
	if (fJoy)
		SDL_JoystickClose(fJoy);
}

bool 
MgJoystickBase::BelongsToUs(Uint8 which) const
{
	return (SDL_JoystickIndex(fJoy) == (int)which) ? true : false;
}

MgJoystickBase::Error 
MgJoystickBase::InitCheck() const
{
	return fError;
}

//===================================================================================

/*! \param joystick The joystick index to open. If you pass -1 as the joystick, the first unopened joystick will be opened. */
MgJoystick::MgJoystick(int joystick)
	: MgEventFilter(), MgJoystickBase(joystick)
{
	if (fError != NoError)
		return;
		
	if (MgApp::App())
	{
		if (!MgApp::App()->AddFilter(this, MG_JOYSTICKEVENT, MG_FILTER_SYSTEM))
			fError = AddFailed;
	}
	else
	{
		fError = AddFailed;
	}
}


MgJoystick::~MgJoystick()
{
	if (fError == NoError)
	{
		if (MgApp::App())
			MgApp::App()->RemoveFilter(this);
	}
}

bool 
MgJoystick::JoystickAxis(const MgJoyAxisEvent & e)
{
	if (BelongsToUs(e.Which()))
	{
		fAxisEvent = e;
		Activated(Axis);
	}
	return true;
}

bool 
MgJoystick::JoystickBall(const MgJoyBallEvent & e)
{
	if (BelongsToUs(e.Which()))
	{
		fBallEvent = e;
		Activated(Ball);
	}
	return true;
}

bool 
MgJoystick::JoystickHat(const MgJoyHatEvent & e)
{
	if (BelongsToUs(e.Which()))
	{
		fHatEvent = e;
		Activated(Hat);
	}
	return true;
}

bool 
MgJoystick::JoystickButton(const MgJoyButtonEvent & e)
{
	if (BelongsToUs(e.Which()))
	{
		fButtonEvent = e;
		Activated(Button);
	}
	return true;
}

//===================================================================================
// Static joystick implementation
//===================================================================================


/*! \param joystick The joystick index to open. If you pass -1 as the joystick, the first unopened joystick will be opened. */
MgSJoystick::MgSJoystick(int joystick)
	: MgJoystickBase(joystick)
{
	if (fError != NoError)
		return;
	
	fStatus = new MgSJoystick::Status;
	if (!fStatus)
	{
		fError = NoMemory;
		return;
	}
	fStatus->fButtons = new Uint8[NumButtons()];
	fStatus->fAxes = new Sint16[NumAxes()];
	fStatus->fHats = new Uint8[NumHats()];
	fStatus->fBalls = new MgSJoystick::Status::Ball[NumBalls()];
	
	if (!fStatus->fButtons || !fStatus->fAxes || !fStatus->fHats || !fStatus->fBalls)
	{
		delete fStatus;
		fError = NoMemory;
		fStatus = NULL;
		return;
	}
}


MgSJoystick::~MgSJoystick()
{
	if (fError == NoError)
		delete fStatus;
}

const MgSJoystick::Status *  
MgSJoystick::GrabStatus()
{
	int i;
	
	/* update buttons */
	for (i = 0; i < NumButtons(); i++)
		fStatus->fButtons[i] = SDL_JoystickGetButton(fJoy, i);
	
	/* update balls */
	for (i = 0; i < NumBalls(); i++)
		SDL_JoystickGetBall(fJoy, i, &fStatus->fBalls[i].dx, &fStatus->fBalls[i].dy);
	
	/* update hats */
	for (i = 0; i < NumHats(); i++)
		fStatus->fHats[i] = SDL_JoystickGetHat(fJoy, i);
		
	/* update axes */
	for (i = 0; i < NumAxes(); i++)
		fStatus->fAxes[i] = SDL_JoystickGetAxis(fJoy, i);
	
	/* return */
	return fStatus;
}

const MgSJoystick::Status *  
MgSJoystick::GetStatus() const
{
	return fStatus;
}
