// MgMutex.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#include "MgMutex.h"

#include "debug_new.h"

/*! Creates the mutex. */
MgMutex::MgMutex()
{
	fMutex = SDL_CreateMutex();
}

/*! Destroys the mutex. */
MgMutex::~MgMutex()
{
	SDL_DestroyMutex(fMutex);
}

/*!
	\retval true Mutex was locked.
	\retval false An error occured. Mutex was not locked!
*/
bool 
MgMutex::Lock()
{
	return (SDL_mutexP(fMutex) == 0) ? true : false;
}

/*!
	\retval true Mutex was unlocked.
	\retval false An error occured!
*/
bool 
MgMutex::Unlock()
{
	return (SDL_mutexV(fMutex) == 0) ? true : false;
}

// ==================================================================================================

/*!
	Creates the semaphore.
	
	\param startVal Initial count of the semaphore.
*/
MgSemaphore::MgSemaphore(Uint32 startVal)
{
	fSem = SDL_CreateSemaphore(startVal);
}

/*! Destroys the semaphore. */
MgSemaphore::~MgSemaphore()
{
	SDL_DestroySemaphore(fSem);
}

/*!
	If the count is positive, it decreases the count atomically and returns,
	otherwise it waits for the count to be positive, then does that.
*/
bool 
MgSemaphore::Lock()
{
	SDL_SemWait(fSem);
	return true;
}

/*!
	\retval true The lock was acquired.
	\retval false The attempt timedout or an error occurred.
*/
bool 
MgSemaphore::TryLock()
{
	return (SDL_SemTryWait(fSem) == 0) ? true : false;
}

/*!
	\param ms The milliseconds to wait if lock can't be acquired.
	
	\retval true Lock was acquired.
	\retval false Lock timedout or there was an error.
*/
bool 
MgSemaphore::Lock(Uint32 ms)
{
	return (SDL_SemWaitTimeout(fSem, ms) == 0) ? true : false;
}

/*!
	\retval true Unlocked.
	\retval false An error occured.
*/
bool 
MgSemaphore::Unlock()
{
	return (SDL_SemPost(fSem) == 0) ? true : false;
}

Uint32 
MgSemaphore::Value()
{
	return SDL_SemValue(fSem);
}

// ==================================================================================================

/*! Creates the condition. */
MgCond::MgCond()
{
	fCond = SDL_CreateCond();
}

/*! Destroys the condition. */
MgCond::~MgCond()
{
	SDL_DestroyCond(fCond);
}

/*!
	\retval true Success.
	\retval false An error occurred.
*/
bool 
MgCond::Signal()
{
	return (SDL_CondSignal(fCond) == 0) ? true : false;
}

/*!
	\retval true Success.
	\retval false An error occured.
*/
bool 
MgCond::Broadcast()
{
	return (SDL_CondBroadcast(fCond) == 0) ? true : false;
}

/*!
	\param mutex The mutex to be used. Do not lock this! This function will do it!
	
	\retval true Calling thread is unlocked because the signal was received.
	\retval false Error!
*/
bool 
MgCond::Wait(MgMutex & mutex)
{
	if (mutex.Lock())
	{
		return (SDL_CondWait(fCond, mutex) == 0) ? true : false;
	}
	return false;
}

/*!
	\param mutex The mutex to be used. Do not lock this! This function will do it!
	\param ms The milliseconds to wait for signal.
	
	\retval true The signal was received.
	\retval false The call timedout or there was an error.
*/
bool 
MgCond::Wait(MgMutex & mutex, Uint32 ms)
{
	if (mutex.Lock())
	{
		return (SDL_CondWaitTimeout(fCond, mutex, ms) == 0) ? true : false;
	}
	return false;
}

