// MgImage.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#include "MgImage.h"
#include "MgMutex.h"
#include "SDL_image.h"
#include "SDL_writeImage.h"
#include "MgPainter.h"

#include "debug_new.h"

#include <stdio.h>

/*! Does nothing. The image is invalid. */
MgImage::MgImage()
	: MgSurface()
{
	fFree = false;
}

/*!
	Loads the image from 'file'.
	
	\param file The image file.
*/
MgImage::MgImage(const char * file)
	: MgSurface()
{
	fFree = false;
	if (LoadImage(file))
	{
		fPixelFormat = new MgPixelFormat(fSurface->format);
		InitSurface();
		fFree = true;
	}
}

/*! Copies the surface and uses it as the image. */
MgImage::MgImage(MgSurface * surface)
	: MgSurface()
{
	fFree = false;
	fSurface = NULL;
	MgPainter * p = surface->ConvertSurface(surface->Format(), SDL_SWSURFACE);
	if (p)
	{
		fSurface = p->fSurface;
		p->FreeOnDelete(false);
		delete p;
		fPixelFormat = new MgPixelFormat(fSurface->format);
		InitSurface();
		fFree = true;
	}
}

/*! Does nothing. MgSurface handles destroying the surface. */
MgImage::~MgImage()
{
}

/*!
	If an image is already loaded, it frees it, and reloads the new image in 'file'

	\param file The image file to load.
	
	\retval true Success.
	\retval false Failure.
*/
bool 
MgImage::Load(const char * file)
{
	if (fPixelFormat)
	{
		delete fPixelFormat;
		fPixelFormat = NULL;
	}
	if (fSurface)
	{
		SDL_FreeSurface(fSurface);
		fSurface = NULL;
	}
	if (fLocker)
	{
		delete fLocker;
		fLocker = NULL;
	}
	fFree = false;
	
	if (LoadImage(file))
	{
		fPixelFormat = new MgPixelFormat(fSurface->format);
		InitSurface();
		fFree = true;
		return true;
	}
	return false;
}

/*!
	\param file The image file to load.
	\param src The source MgSurface to save.
	
	\retval true Image was saved.
	\retval false Failed to save the image.
*/
bool 
MgImage::SaveJPG(const char * file, MgSurface * src)
{
	bool ret = false;
	if (src->Lock())
	{
		FILE * fp = fopen(file, "wb");
		if (!fp)
		{
			src->Unlock();
			return false;
		}
		ret = IMG_SaveJPG(fp, src) == 0 ? true : false;
		fclose(fp);
		src->Unlock();
	}
	return ret;
}

/*!
	\param file The image file to load.
	\param src The source MgSurface to save.
	
	\retval true Image was saved.
	\retval false Failed to save the image.
*/
bool 
MgImage::SavePNG(const char * file, MgSurface * src)
{
	bool ret = false;
	if (src->Lock())
	{
		FILE * fp = fopen(file, "wb");
		if (!fp)
		{
			src->Unlock();
			return false;
		}
		ret = IMG_SavePNG(fp, src) == 0 ? true : false;
		fclose(fp);
		src->Unlock();
	}
	return ret;
}

/*!
	\param file The image file to load.
	\param src The source MgSurface to save.
	
	\retval true Image was saved.
	\retval false Failed to save the image.
*/
bool 
MgImage::SavePPM(const char * file, MgSurface * src)
{
	bool ret = false;
	if (src->Lock())
	{
		ret = IMG_SavePPM(file, src, 1) == 0 ? true : false;
		src->Unlock();
	}
	return ret;
}

/*!
	\param file The image file to load.
	\param src The source MgSurface to save.
	
	\retval true Image was saved.
	\retval false Failed to save the image.
*/
bool 
MgImage::SaveBMP(const char * file, MgSurface * src)
{
	bool ret = false;
	if (src->Lock())
	{
		ret = SDL_SaveBMP(src->fSurface, file) == 0 ? true : false;
		src->Unlock();
	}
	return ret;
}

void 
MgImage::FreeOnDelete(bool free)
{
	// always free if valid
	if (InitCheck())
		fFree = true;
	else
		fFree = false;
}

bool
MgImage::InitCheck() const
{
	return (fFree && MgSurface::InitCheck()) ? true : false;
}

// interal loading function
bool 
MgImage::LoadImage(const char * file)
{
	fSurface = IMG_Load(file);
	
	if (!fSurface)
		return false;
	return true;
}

