// MgSurface.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#include "MgSurface.h"
#include "MgPainter.h"
#include "MgPixelFormat.h"
#include "MgPalette.h"
#include "MgMutex.h"
#include "internal/MgDraw.h"
#include "debug_new.h"

//#include <stdio.h>

// to avoid a function called when locking
#define LOCK(p) p->fLocker->Lock()
#define UNLOCK(p) p->fLocker->Unlock()

// inline function to lock HW surfaces
#define LOCKHW(p) SDL_MUSTLOCK(p->fSurface) ? ((SDL_LockSurface(p->fSurface) == 0) ? true : false) : true	// to be used in "if" statements
#define UNLOCKHW(p) if (SDL_MUSTLOCK(p->fSurface)) SDL_UnlockSurface(p->fSurface)

/*! Does nothing. The object is invalid if this constructor is used. */
MgSurface::MgSurface()
{
	fPixelFormat = NULL;
	fSurface = NULL;
	fFree = false;
}

/*! Creates the surface from 'surface' */
MgSurface::MgSurface(SDL_Surface * surface)
{
	fSurface = surface;
	fPixelFormat = new MgPixelFormat(surface->format);
	InitSurface();
	fFree = false;
}

/*! Creates the surface from the passed vars. */
MgSurface::MgSurface(Uint32 flags, int width, int height, int depth, Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask)
{
	fSurface = SDL_CreateRGBSurface(flags, width, height, depth, Rmask, Gmask, Bmask, Amask);
	if (fSurface)
		fPixelFormat = new MgPixelFormat(fSurface->format);
	InitSurface();
	fFree = false;
}

/*! Creates the surface with the content in 'pixels' from the passed vars. */
MgSurface::MgSurface(void * pixels, int width, int height, int depth, int pitch, Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask)
{
	fSurface = SDL_CreateRGBSurfaceFrom(pixels, width, height, depth, pitch, Rmask, Gmask, Bmask, Amask);
	if (fSurface)
		fPixelFormat = new MgPixelFormat(fSurface->format);
	InitSurface();
	fFree = false;
}

/*! Destroys the surface. */
MgSurface::~MgSurface()
{
//	printf("Destroy MgSurface\n");
	
	if (fFree)
		Free();
	delete fPixelFormat;	// freed no matter what
	delete fLocker;
}

MgPixelFormat * 
MgSurface::Format() const
{
	return fPixelFormat;
}

Uint32 
MgSurface::Flags() const
{
	return fSurface->flags;
}

/*!
	Be aware that if you use this pointer to draw onto the surface,
	it is not locked! So another thread could do the same.
*/
Uint8 * 
MgSurface::Pixels() const
{
	return (Uint8 *)fSurface->pixels;
}

int 
MgSurface::Width() const
{
	return fSurface->w;
}

int 
MgSurface::Height() const
{
	return fSurface->h;
}

Uint16 
MgSurface::Pitch() const
{
	return fSurface->pitch;
}

MgRect 
MgSurface::ClipRect() const
{
	SDL_Rect rect = { 0, 0, 0, 0 };
	if (LOCK(this))
	{
		SDL_GetClipRect(fSurface, &rect);
		UNLOCK(this);
	}
	return MgRect(rect);
}

void 
MgSurface::Clear(MgColor col)
{
	if (LOCK(this))
	{
		Uint32 color = fPixelFormat->GetPixel(col);
		SDL_FillRect(fSurface, NULL, color);
		UNLOCK(this);
	}
}

/*!
	This locking function does not do the same thing as LockHardware()! This function
	locks the surface for use by MgSurface's drawing/blit functions. Therefore making
	MgSurface's thread safe. If you use one of the Blit() functions and you pass an
	SDL_Surface to it instead of a MgSurface, the SDL_Surface is not locked! But
	if you use the MgSurface version, that surface will be locked!
	
	\retval true Locked.
	\retval false Failed to lock.
*/
bool 
MgSurface::Lock()
{
	return fLocker->Lock();
}

/*!
	\retval true Unlocked.
	\retval false Error!
*/
bool 
MgSurface::Unlock()
{
	return fLocker->Unlock();
}

/*!
	This function will lock a hardware surface for direct access to the pixel data.
	
	\retval true Locked or didn't need to lock.
	\retval false Lock failed.
*/
bool 
MgSurface::LockHardware()
{
	if (SDL_MUSTLOCK(fSurface))
	{
		return (SDL_LockSurface(fSurface) == 0) ? true : false;
	}
	return true;		// don't need to..
}

/*! Always returns true. */
bool 
MgSurface::UnlockHardware()
{
	if (SDL_MUSTLOCK(fSurface))
		SDL_UnlockSurface(fSurface);
	return true;	// 
}

void 
MgSurface::UpdateRect(Sint32 x, Sint32 y, Uint32 w, Uint32 h)
{
	if(LOCK(this))
	{
		SDL_UpdateRect(fSurface, x, y, w, h);
		UNLOCK(this);
	}
}

void 
MgSurface::UpdateRects(SDL_Rect * rects, int numRects)
{
	if (LOCK(this))
	{
		SDL_UpdateRects(fSurface, numRects, rects);
		UNLOCK(this);
	}
}

void 
MgSurface::UpdateRects(MgRect * rects, int numRects)
{
	SDL_Rect * rct = new SDL_Rect[numRects];
	
	if (LOCK(this))
	{
		for (int i = 0; i < numRects; i++)
			rct[i] = rects[i];
		SDL_UpdateRects(fSurface, numRects, rct);
		UNLOCK(this);
	}
	delete [] rct;
}

bool 
MgSurface::ApplyPalette()
{
	if (LOCK(this))
	{
		if (fPixelFormat->BytesPerPixel() == 1)
		{
			fPixelFormat->Palette()->ApplyTo(this);
			UNLOCK(this);
			return true;
		}
		UNLOCK(this);
	}
	return false;	// not paletted
}

bool 
MgSurface::SetColorKey(Uint32 flag, Uint32 key)
{
	bool ret = false;
	
	if (LOCK(this))
	{
		ret = (SDL_SetColorKey(fSurface, flag, key) == 0) ? true : false;
		UNLOCK(this);
	}
	return ret;
}

bool 
MgSurface::SetAlpha(Uint32 flag, Uint8 alpha)
{
	bool ret = false;
	
	if (LOCK(this))
	{
		ret = (SDL_SetAlpha(fSurface, flag, alpha) == 0) ? true : false;
		UNLOCK(this);
	}
	return ret;
}

bool 
MgSurface::SetClipRect(SDL_Rect & rect)
{
	bool ret = false;
	
	if (LOCK(this))
	{
		ret = (SDL_SetClipRect(fSurface, &rect) == 0) ? true : false;
		UNLOCK(this);
	}
	return ret;
}

bool 
MgSurface::SetClipRect(MgRect & rect)
{
	SDL_Rect rct = rect;
	bool ret = false;

	if (LOCK(this))
	{
		ret = (SDL_SetClipRect(fSurface, &rct) == 0) ? true : false;
		UNLOCK(this);
	}
	return ret;
}

/*!
	\return If NULL, the the operation failed. Otherwise it was successful and returns a pointer to a new MgPainter. You must free it yourself.
*/
MgPainter * 
MgSurface::ConvertSurface(SDL_PixelFormat * fmt, Uint32 flags)
{
	SDL_Surface * surface = NULL;
	MgPainter * painter = NULL;
	
	if (LOCK(this))
	{
		surface = SDL_ConvertSurface(fSurface, fmt, flags);
		if (surface)
			painter = new MgPainter(surface);
		UNLOCK(this);
	}
	return painter;
}

/*!
	\return If NULL, the the operation failed. Otherwise it was successful and returns a pointer to a new MgPainter. You must free it yourself.
*/
MgPainter * 
MgSurface::ConvertSurface(MgPixelFormat * fmt, Uint32 flags)
{
	SDL_Surface * surface = NULL;
	MgPainter * painter = NULL;

	if (LOCK(this))
	{
		surface = SDL_ConvertSurface(fSurface, fmt->fFormat, flags);
		if (surface)
			painter = new MgPainter(surface);
		UNLOCK(this);
	}
	return painter;
}

/*!
	\retval NULL An error occurred.
	\retval Non-NULL A valid MgPainter pointer. You must free the painter yourself.
*/
MgPainter * 
MgSurface::DisplayFormat()
{
	SDL_Surface * surface = NULL;
	MgPainter * painter = NULL;

	if (LOCK(this))
	{
		surface = SDL_DisplayFormat(fSurface);
		if (surface)
			painter = new MgPainter(surface);
		UNLOCK(this);
	}
	return painter;
}

/*!
	\retval NULL Error!
	\retval Non-NULL A valid MgPainter pointer.
*/
MgPainter * 
MgSurface::DisplayFormatAlpha()
{
	SDL_Surface * surface = NULL;
	MgPainter * painter = NULL;

	if (LOCK(this))
	{
		surface = SDL_DisplayFormatAlpha(fSurface);
		if (surface)
			painter = new MgPainter(surface);
		UNLOCK(this);
	}
	return painter;	
}

void 
MgSurface::FreeOnDelete(bool free)
{
	fFree = free;
}

/*!
	Blit the whole source surface 'src' to the 'x' and 'y' points on this surface.
	
	\param src Source surface.
	\param x X destination.
	\param y Y destination.
	
	\retval true Success.
	\retval false Failure.
*/
bool 
MgSurface::Blit(MgSurface * src, int x, int y)
{
	SDL_Rect srcrect, dstrect;
	bool ret = false;
	
	if (LOCK(this))
	{
		if (LOCK(src))
		{
			srcrect.x = 0;
			srcrect.y = 0;
			dstrect.x = x;
			dstrect.y = y;
			srcrect.w = dstrect.w = src->Width();
			srcrect.h = dstrect.h = src->Height();
	
			ret = (SDL_BlitSurface(src->fSurface, &srcrect, fSurface, &dstrect) == 0) ? true : false;
			UNLOCK(src);
		}
		UNLOCK(this);
	}
	return ret;
}

/*!
	Blit the whole source surface 'src' to the 'x' and 'y' points on this surface.
	
	\param src Source surface.
	\param x X destination.
	\param y Y destination.
	
	\retval true Success.
	\retval false Failure.
*/
bool 
MgSurface::Blit(SDL_Surface * src, int x, int y)
{
	SDL_Rect srcrect, dstrect;
	bool ret = false;
	
	if (LOCK(this))
	{
		srcrect.x = 0;
		srcrect.y = 0;
		dstrect.x = x;
		dstrect.y = y;
		srcrect.w = dstrect.w = src->w;
		srcrect.h = dstrect.h = src->h;

		ret = (SDL_BlitSurface(src, &srcrect, fSurface, &dstrect) == 0) ? true : false;

		UNLOCK(this);
	}
	return ret;
}

/*!
	Blit the 'rect' part of the source surface to the 'rect' position in the destination surface.
	This is useful if you have two surfaces of the same size. You can pass an MgRect for the
	'rect' param (that's what the operator is for in MgRect).
	
	\param src Source surface.
	\param rect The source and destination rectangle.
	
	\retval true Success.
	\retval false Error.
*/
bool 
MgSurface::Blit(MgSurface * src, SDL_Rect rect)
{
	SDL_Rect srcrect, dstrect;
	bool ret = false;
	
	if (LOCK(this))
	{
		if (LOCK(src))
		{
			srcrect.x = dstrect.x = rect.x;
			srcrect.y = dstrect.y = rect.y;
			srcrect.w = dstrect.w = rect.w;
			srcrect.h = dstrect.h = rect.h;
	
			ret = (SDL_BlitSurface(src->fSurface, &srcrect, fSurface, &dstrect) == 0) ? true : false;
			UNLOCK(src);
		}
		UNLOCK(this);
	}
	return ret;
}

/*!
	Blit the 'rect' part of the source surface to the 'rect' position in the destination surface.
	This is useful if you have two surfaces of the same size. You can pass an MgRect for the
	'rect' param (that's what the operator is for in MgRect).
	
	\param src Source surface.
	\param rect The source and destination rectangle.
	
	\retval true Success.
	\retval false Error.
*/
bool 
MgSurface::Blit(SDL_Surface * src, SDL_Rect rect)
{
	SDL_Rect srcrect, dstrect;
	bool ret = false;
	
	if (LOCK(this))
	{
		srcrect.x = dstrect.x = rect.x;
		srcrect.y = dstrect.y = rect.y;
		srcrect.w = dstrect.w = rect.w;
		srcrect.h = dstrect.h = rect.h;
	
		ret = (SDL_BlitSurface(src, &srcrect, fSurface, &dstrect) == 0) ? true : false;

		UNLOCK(this);
	}
	return ret;
}

/*!
	Blit the 'srcrect' part of the source surface 'src' to the 'dstrect' on the destination surface. You can
	pass an MgRect for the 'srcrect' and 'dstrect' parameters.
	
	\param src Source surface.
	\param srcrect Source rect.
	\param dstrect Destination rect.
	
	\retval true Success.
	\retval false Error!
*/
bool 
MgSurface::Blit(MgSurface * src, SDL_Rect srcrect, SDL_Rect dstrect)
{
	bool ret = false;
	
	if (LOCK(this))
	{
		if (LOCK(src))
		{
			ret = (SDL_BlitSurface(src->fSurface, &srcrect, fSurface, &dstrect) == 0) ? true : false;
			UNLOCK(src);
		}
		UNLOCK(this);
	}
	return ret;
		
}

/*!
	Blit the 'srcrect' part of the source surface 'src' to the 'dstrect' on the destination surface. You can
	pass an MgRect for the 'srcrect' and 'dstrect' parameters.
	
	\param src Source surface.
	\param srcrect Source rect.
	\param dstrect Destination rect.
	
	\retval true Success.
	\retval false Error!
*/
bool 
MgSurface::Blit(SDL_Surface * src, SDL_Rect srcrect, SDL_Rect dstrect)
{
	bool ret = false;
	
	if (LOCK(this))
	{
		ret = (SDL_BlitSurface(src, &srcrect, fSurface, &dstrect) == 0) ? true : false;
		UNLOCK(this);
	}
	return ret;
}

/*!
	Blit the source surface 'src' from 'srcrect' onto this surface at 'x' and 'y'.
	
	\param src Source surface.
	\param srcrect Rectangle of source surface to be blitted.
	\param x X destination.
	\param y Y destination.
	
	\retval true Success.
	\retval false Error.
*/
bool 
MgSurface::Blit(MgSurface * src, SDL_Rect srcrect, int x, int y)
{
	SDL_Rect dstrect;
	bool ret = false;
	
	if (LOCK(this))
	{
		if (LOCK(src))
		{
			dstrect.x = x;
			dstrect.y = y;
			dstrect.h = srcrect.h;
			dstrect.w = srcrect.w;
			
			ret = (SDL_BlitSurface(src->fSurface, &srcrect, fSurface, &dstrect) == 0) ? true : false;
			UNLOCK(src);
		}
		UNLOCK(this);
	}
	return ret;
}

/*!
	Blit the source surface 'src' from 'srcrect' onto this surface at 'x' and 'y'.
	
	\param src Source surface.
	\param srcrect Rectangle of source surface to be blitted.
	\param x X destination.
	\param y Y destination.
	
	\retval true Success.
	\retval false Error.
*/
bool 
MgSurface::Blit(SDL_Surface * src, SDL_Rect srcrect, int x, int y)
{
	SDL_Rect dstrect;
	bool ret = false;
	
	if (LOCK(this))
	{
		dstrect.x = x;
		dstrect.y = y;
		dstrect.h = srcrect.h;
		dstrect.w = srcrect.w;
		
		ret = (SDL_BlitSurface(src, &srcrect, fSurface, &dstrect) == 0) ? true : false;

		UNLOCK(this);
	}
	return ret;
}

/*!
	\retval true Everything is fine.
	\retval false The object is invalid.
*/
bool 
MgSurface::InitCheck() const
{
	if (fPixelFormat == NULL || fSurface == NULL)
		return false;
	return true;
}

/*! Get a pixel.

	\param x X coord.
	\param y Y coord.
	
	\return The color of the pixel.
*/
Uint32 
MgSurface::GetPixel(int x, int y)
{
	Uint32 ret = 0;
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			ret = getpixel(fSurface, x, y);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
	
	return ret;
}

//@{
/*! Draw a pixel.
	
	\param x X coord.
	\param y Y coord.
	\param col Color to draw it in.
*/
void 
MgSurface::PutPixel(int x, int y, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			putpixel(fSurface, x, y, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

void
MgSurface::PutPixel(int x, int y, Uint32 col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			putpixel2(fSurface, x, y, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}
//@}

/*! Draw a horizontal line.
	
	\param x1 Left x position.
	\param x2 Right x position.
	\param y Y position.
	\param col Color to draw it in.
*/
void 
MgSurface::StrokeHLine(int x1, int x2, int y, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			strokehline(fSurface, x1, x2, y, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a vertical line.

	\param y1 Top y position.
	\param y2 Bottom y position.
	\param x X position.
	\param col Color.
*/
void 
MgSurface::StrokeVLine(int y1, int y2, int x, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			strokevline(fSurface, y1, y2, x, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a line.

	\param x1 Left position.
	\param y1 Top position.
	\param x2 Right position.
	\param y2 Bottom position.
	\param col Color.
*/
void 
MgSurface::StrokeLine(int x1, int y1, int x2, int y2, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			strokeline(fSurface, x1, y1, x2, y2, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a rectangular outline.

	\param x1 Left position.
	\param y1 Top position.
	\param x2 Right position.
	\param y2 Bottom position.
	\param col Color.
*/
void 
MgSurface::StrokeRect(int x1, int y1, int x2, int y2, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			strokerect(fSurface, x1, y1, x2, y2, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a circular outline.

	\param x X position.
	\param y Y position.
	\param radius The radius.
	\param col Color.
*/
void 
MgSurface::StrokeCircle(int x, int y, int radius, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			strokecircle(fSurface, x, y, radius, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw an outline of an ellipse.

	\param x X position.
	\param y Y position.
	\param xr X radius.
	\param yr T radius.
	\param col Color.
*/
void 
MgSurface::StrokeEllipse(int x, int y, int xr, int yr, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			strokeellipse(fSurface, x, y, xr, yr, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw an outline of a triangle

	\param x1/y1/x2/y2/x3/y3 The vertices of the triangle.
	\param col Color.
*/
void 
MgSurface::StrokeTriangle(int x1, int y1, int x2, int y2, int x3, int y3, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			stroketriangle(fSurface, x1, y1, x2, y2, x3, y3, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw an outline of a polygon. As you notice, the MgVertex class has a
	color component. This is used in this method to draw the lines. Therefore
	you can have a polygon with different colored lines or with just one color.
	
	\param vertices Vertices of the polygon.
	\param numVertices Number of vertices. Therefore (numVertices - 1) is upperbounds of 'vertices'.
*/
void 
MgSurface::StrokePolygon(MgIVertex * vertices, int numVertices)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			strokepolygon(fSurface, vertices, numVertices);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a filled rectangle.

	\param x1 Left position.
	\param y1 Top position.
	\param x2 Right position.
	\param y2 Bottom position.
	\param col Color.
*/
void 
MgSurface::FillRect(int x1, int y1, int x2, int y2, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			fillrect(fSurface, x1, y1, x2, y2, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a filled circle.
	
	\param x X position.
	\param y Y position.
	\param radius Radius.
	\param col Color.
*/
void 
MgSurface::FillCircle(int x, int y, int radius, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			fillcircle(fSurface, x, y, radius, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a filled ellipse.

	\param x X position.
	\param y Y position.
	\param xr X radius.
	\param yr Y radius.
	\param col Color.
*/
void 
MgSurface::FillEllipse(int x, int y, int xr, int yr, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			fillellipse(fSurface, x, y, xr, yr, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a filled triangle

	\param x1/y1/x2/y2/x3/y3 The vertices of the triangle.
	\param col Color.
*/
void 
MgSurface::FillTriangle(int x1, int y1, int x2, int y2, int x3, int y3, MgColor col)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			filltriangle(fSurface, x1, y1, x2, y2, x3, y3, col);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a filled polygon. This version has only one color! It does not pay any attention to the
	colors in MgIVertex. It works with concave and convex polygons.
	
	\param vertices An array of vertices.
	\param n Number of vertices so that "n - 1" will be the last element of the array.
	\param color Color of the polygon.
*/
void 
MgSurface::FillPolygon(MgIVertex * vertices, int n, MgColor color)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			fillpolygon(fSurface, vertices, n, color);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a gradient blending the top color with the bottom color.
	
	\param x1 Left position.
	\param y1 Top position.
	\param x2 Right position.
	\param y2 Bottom position.
	
	\param startCol Top color.
	\param endCol The bottom color to blend to.
*/
void 
MgSurface::FillGradient(int x1, int y1, int x2, int y2, MgColor startCol, MgColor endCol)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			fillgradient(fSurface, x1, y1, x2, y2, startCol, endCol);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a gouraud shaded horizontal line.

	\param x1 Left position.
	\param x2 Right position.
	\param y Y position.
	\param c1 Left color.
	\param c2 Right color.
*/
void 
MgSurface::ShadeHLine(int x1, int x2, int y, MgColor c1, MgColor c2)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			shadehline(fSurface, x1, x2, y, c1, c2);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a gouraud shaded vertical line.
	
	\param y1 Top position.
	\param y2 Bottom position.
	\param x X position.
	\param c1 Top color.
	\param c2 Bottom color.
*/
void
MgSurface::ShadeVLine(int y1, int y2, int x, MgColor c1, MgColor c2)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			shadevline(fSurface, y1, y2, x, c1, c2);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a gouraud shaded line.

	\param x1 Left position.
	\param y1 Top position.
	\param x2 Right position.
	\param y2 Bottom position.
	\param c1 Top-left color.
	\param c2 Bottom-right color.
*/
void 
MgSurface::ShadeLine(int x1, int y1, int x2, int y2, MgColor c1, MgColor c2)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			shadeline(fSurface, x1, y1, x2, y2, c1, c2);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a gouraud shaded rect.

	\param x1 Left point.
	\param y1 Top point.
	\param x2 Right point.
	\param y2 Bottom point.
	\param ul Upper-left color.
	\param ur Upper-right color.
	\param dl Lower-left color.
	\param dr Lower-right color.
*/
void 
MgSurface::ShadeRect(int x1, int y1, int x2, int y2, MgColor ul, MgColor ur, MgColor dl, MgColor dr)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			shaderect(fSurface, x1, y1, x2, y2, ul, ur, dl, dr);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a gouraud shaded circle blending a color from the center to the destination color at
	the outer edge of the circle.
	
	\param x X position.
	\param y Y position.
	\param center Center color.
	\param outside Color at the edge.
*/
void 
MgSurface::ShadeCircle(int x, int y, int r, MgColor center, MgColor outside)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			shadecircle(fSurface, x, y, r, center, outside);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a gouraud shaded triangle.

	\param x1/y1/x2/y2/x3/y3 The vertices of the triangle.
	\param cl/c2/c3 The respective colors of the vertices.
*/
void 
MgSurface::ShadeTriangle(int x1, int y1, int x2, int y2, int x3, int y3, MgColor c1, MgColor c2, MgColor c3)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			shadetriangle(fSurface, x1, y1, x2, y2, x3, y3, c1, c2, c3);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draw a gouraud shaded polygon. Be sure to note that this only works with <b>concave</b>
	polygons! The color component of MgIVertex is the color for the vertex.
	
	\param vertices The vertices.
	\param numVertices Number of vertices so that "numVertices - 1" is the last element in the array.
*/
void 
MgSurface::ShadePolygon(MgIVertex * vertices, int numVertices)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			shadepolygon(fSurface, vertices, numVertices);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draws an unfilled circle blending the colors for the "left", "right", "bottom", and "top"
	parts of the circle. Play with the function and see what I mean.
	
	\param x X position.
	\param y Y position.
	\param top Top color.
	\param right Right color.
	\param bottom Bottom color.
	\param left Left color.
*/
void 
MgSurface::ShadeCircleOutline(int x, int y, int r, MgColor top, MgColor right, MgColor bottom,
								MgColor left)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			shadecircleoutline(fSurface, x, y, r, top, right, bottom, left);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

/*! Draws an unfilled rectangle blending the colors at the vertices.

	\param x1 Left position.
	\param y1 Top position.
	\param x2 Right position.
	\param y2 Bottom position.
	\param topleft Top-left vertex color.
	\param topright Top-right vertex color.
	\param bottomright Bottom-right vertex color.
	\param bottomleft Bottom-left vertex color.
*/
void 
MgSurface::ShadeRectOutline(int x1, int y1, int x2, int y2, MgColor topleft, MgColor topright,
							MgColor	bottomright, MgColor bottomleft)
{
	if (LOCK(this))
	{
		if (LOCKHW(this))
		{
			shaderectoutline(fSurface, x1, y1, x2, y2, topleft, topright, bottomright, bottomleft);
			UNLOCKHW(this);
		}
		UNLOCK(this);
	}
}

void 
MgSurface::Free()
{
//	printf("freeing\n");
	if (LOCK(this))
	{
		SDL_FreeSurface(fSurface);
		UNLOCK(this);
	}
}

void 
MgSurface::InitSurface()
{
	fLocker = new MgMutex();
	switch (fPixelFormat->BytesPerPixel())
	{
		case 1:
		{
			getpixel = GetPixel8;
			putpixel = PutPixel8;
			putpixel2 = PutPixel8;
			strokehline = StrokeHLine8;
			strokevline = StrokeVLine8;
			strokeline = StrokeLine8;
			strokerect = StrokeRect8;
			strokecircle = StrokeCircle8;
			strokeellipse = StrokeEllipse8;
			stroketriangle = StrokeTriangle8;
			strokepolygon = StrokePolygon8;
			fillrect = FillRect8;
			fillcircle = FillCircle8;
			fillellipse = FillEllipse8;
			filltriangle = FillTriangle8;
			fillpolygon = FillPolygon8;
			fillgradient = FillGradient8;
			shadehline = ShadeHLine8;
			shadevline = ShadeVLine8;
			shadeline = ShadeLine8;
			shaderect = ShadeRect8;
			shadecircle = ShadeCircle8;
			shadetriangle = ShadeTriangle8;
			shadepolygon = ShadePolygon8;
			shadecircleoutline = ShadeCircleOutline8;
			shaderectoutline = ShadeRectOutline8;
			break;
		}
		
		case 2:
		{
			getpixel = GetPixel16;
			putpixel = PutPixel16;
			putpixel2 = PutPixel16;
			strokehline = StrokeHLine16;
			strokevline = StrokeVLine16;
			strokeline = StrokeLine16;
			strokerect = StrokeRect16;
			strokecircle = StrokeCircle16;
			strokeellipse = StrokeEllipse16;
			stroketriangle = StrokeTriangle16;
			strokepolygon = StrokePolygon16;
			fillrect = FillRect16;
			fillcircle = FillCircle16;
			fillellipse = FillEllipse16;
			filltriangle = FillTriangle16;
			fillpolygon = FillPolygon16;
			fillgradient = FillGradient16;
			shadehline = ShadeHLine16;
			shadevline = ShadeVLine16;
			shadeline = ShadeLine16;
			shaderect = ShadeRect16;
			shadecircle = ShadeCircle16;
			shadetriangle = ShadeTriangle16;
			shadepolygon = ShadePolygon16;
			shadecircleoutline = ShadeCircleOutline16;
			shaderectoutline = ShadeRectOutline16;
			break;
		}
		
		case 3:
		{
			getpixel = GetPixel24;
			putpixel = PutPixel24;
			putpixel2 = PutPixel24;
			strokehline = StrokeHLine24;
			strokevline = StrokeVLine24;
			strokeline = StrokeLine24;
			strokerect = StrokeRect24;
			strokecircle = StrokeCircle24;
			strokeellipse = StrokeEllipse24;
			stroketriangle = StrokeTriangle24;
			strokepolygon = StrokePolygon24;
			fillrect = FillRect24;
			fillcircle = FillCircle24;
			fillellipse = FillEllipse24;
			filltriangle = FillTriangle24;
			fillpolygon = FillPolygon24;
			fillgradient = FillGradient24;
			shadehline = ShadeHLine24;
			shadevline = ShadeVLine24;
			shadeline = ShadeLine24;
			shaderect = ShadeRect24;
			shadecircle = ShadeCircle24;
			shadetriangle = ShadeTriangle24;
			shadepolygon = ShadePolygon24;
			shadecircleoutline = ShadeCircleOutline24;
			shaderectoutline = ShadeRectOutline24;
			break;
		}

		case 4:
		{
			getpixel = GetPixel32;
			putpixel = PutPixel32;
			putpixel2 = PutPixel32;
			strokehline = StrokeHLine32;
			strokevline = StrokeVLine32;
			strokeline = StrokeLine32;
			strokerect = StrokeRect32;
			strokecircle = StrokeCircle32;
			strokeellipse = StrokeEllipse32;
			stroketriangle = StrokeTriangle32;
			strokepolygon = StrokePolygon32;
			fillrect = FillRect32;
			fillcircle = FillCircle32;
			fillellipse = FillEllipse32;
			filltriangle = FillTriangle32;
			fillpolygon = FillPolygon32;
			fillgradient = FillGradient32;
			shadehline = ShadeHLine32;
			shadevline = ShadeVLine32;
			shadeline = ShadeLine32;
			shaderect = ShadeRect32;
			shadecircle = ShadeCircle32;
			shadetriangle = ShadeTriangle32;
			shadepolygon = ShadePolygon32;
			shadecircleoutline = ShadeCircleOutline32;
			shaderectoutline = ShadeRectOutline32;
			break;
		}
	}
}

