// MgVideoSystem.cpp
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#include "MgVideoSystem.h"
#include "MgPixelFormat.h"
#include "MgPalette.h"
#include "MgScreen.h"
#include "MgGLScreen.h"
#include "MgApp.h"
#include "debug_new.h"

#include <SDL.h>
#include <string.h>

MgVideoInfo::MgVideoInfo(const SDL_VideoInfo * info)
{
	fInfo = info;
	fFormat = new MgPixelFormat(fInfo->vfmt);
}


MgVideoInfo::~MgVideoInfo()
{
	delete fFormat;
}

bool 
MgVideoInfo::HWSurface() const
{
	return fInfo->hw_available ? true : false;
}

bool 
MgVideoInfo::WindowManager() const
{
	return fInfo->wm_available ? true : false;
}

bool 
MgVideoInfo::HWBlit() const
{
	return fInfo->blit_hw ? true : false;
}

bool 
MgVideoInfo::HWColorKeyBlit() const
{
	return fInfo->blit_hw_CC ? true : false;
}

bool 
MgVideoInfo::HWAlphaBlit() const
{
	return fInfo->blit_hw_A ? true : false;
}

bool 
MgVideoInfo::SWBlit() const
{
	return fInfo->blit_sw ? true : false;
}

bool 
MgVideoInfo::SWColorKeyBlit() const
{
	return fInfo->blit_sw_CC ? true : false;
}

bool 
MgVideoInfo::SWAlphaBlit() const
{
	return fInfo->blit_sw_A ? true : false;
}

bool 
MgVideoInfo::FillBlit() const
{
	return fInfo->blit_fill ? true : false;
}

Uint32 
MgVideoInfo::VideoMemoryKB() const
{
	return fInfo->video_mem;
}

Uint32 
MgVideoInfo::VideoMemoryMB() const
{
	return fInfo->video_mem / 1024;	// K -> M conversion
}

const MgPixelFormat *
MgVideoInfo::Format() const
{
	return fFormat;
}


//===================================================================================



MgVideoSystem::MgVideoSystem(const char * title)
{
	fVideoInfo = NULL;
	fScreen = NULL;
	fTitle = NULL;
	fGLScreen = NULL;
	fGL = 0;
	
	if (title)
	{
		fTitle = new char[strlen(title) + 1];
		strcpy(fTitle, title);
	}
}

MgVideoSystem::~MgVideoSystem()
{
	delete fVideoInfo;
	if (fScreen)
		delete fScreen;
	if (fGLScreen)
		delete fGLScreen;
	delete [] fTitle;
}

bool 
MgVideoSystem::IsRunning() const
{
	if (SDL_WasInit(MG_VIDEO) == MG_VIDEO)
		return true;
	return false;
}

bool 
MgVideoSystem::Init()
{
	if (!IsRunning())
	{
		if (SDL_InitSubSystem(MG_VIDEO) < 0)
			return false;
	}
	return true;
}

void 
MgVideoSystem::Quit()
{
	if (IsRunning())
	{
		delete fScreen;
		fScreen = NULL;
		SDL_QuitSubSystem(MG_VIDEO);
	}
}

/*!
	Do not try to create an OpenGL surface with this method, it will fail!
	
	\return If the video mode is not set, it sets it and returns the new MgScreen. If the video mode is set, it returns the MgScreen without resetting the mode.
*/
MgScreen *
MgVideoSystem::SetVideoMode(int width, int height, int depth, Uint32 flags)
{
	if (flags & MG_OPENGL)
		return NULL;
	
	if (fGL)	/* already started for OpenGL use? */
		return NULL;
		
	if (fScreen)
	{
		delete fScreen;
		fScreen = NULL;
	}
	
	SDL_Surface * sur = NULL;
	
	sur = SDL_SetVideoMode(width, height, depth, flags);
	if (!sur)
		return NULL;
	fScreen = new MgScreen(sur);
	
	if (fTitle)
		SDL_WM_SetCaption(fTitle, fTitle);
		
	return fScreen;
}

/*!
	\return If successful, a valid pointer. Otherwise, NULL.
*/
MgGLScreen *
MgVideoSystem::SetGLVideoMode(int width, int height, int depth, Uint32 flags)
{
	if (fScreen)
		return NULL;
	
	if (fGLScreen)	/* already set? */
		return fGLScreen;
	
	flags |= MG_OPENGL;	/* make sure the OpenGL flag is set */
	
	SDL_Surface * sur = NULL;
	
	sur = SDL_SetVideoMode(width, height, depth, flags);
	
	if (!sur)
		return NULL;
	fGLScreen = new MgGLScreen();
	if (!fGLScreen)
		return NULL;
	fGLScreen->fSurface = sur;
	fGL = 1;
	
	if (fTitle)
		SDL_WM_SetCaption(fTitle, fTitle);
	
	return fGLScreen;
}

const MgVideoInfo *
MgVideoSystem::GetVideoInfo()
{
	if (fVideoInfo)
	{
		delete fVideoInfo;
		fVideoInfo = NULL;
	}
	const SDL_VideoInfo * info = SDL_GetVideoInfo();
	if (info)
		fVideoInfo = new MgVideoInfo(info);
	return fVideoInfo;
}

/*!
	\retval 0 The video most is OK.
	\retval Non-0 The closest bits-per-pixel to the requested mode.
*/
int 
MgVideoSystem::VideoModeOK(int width, int height, int bpp, Uint32 flags)
{
	return SDL_VideoModeOK(width, height, bpp, flags);
}

SDL_Rect **
MgVideoSystem::ListModes(MgPixelFormat * format, Uint32 flags)
{
	return SDL_ListModes(format->fFormat, flags);
}

SDL_Rect **
MgVideoSystem::ListModes(SDL_PixelFormat * format, Uint32 flags)
{
	return SDL_ListModes(format, flags);
}

const char *
MgVideoSystem::Title() const
{
	return fTitle;
}

void
MgVideoSystem::SetTitle(const char * title)
{
	delete [] fTitle;
	fTitle = NULL;
	
	if (title)
	{
		fTitle = new char[strlen(title) + 1];
		strcpy(fTitle, title);
	}
	if (fScreen)
		SDL_WM_SetCaption(fTitle, fTitle);
}

/*! The gamma value range is between 0.1 and 10.0

	\retval true Success.
	\retval false Error.
*/
bool 
MgVideoSystem::SetGamma(float red, float green, float blue)
{
	return (SDL_SetGamma(red, green, blue) == 0) ? true : false;
}

/*! Each table is an array of 256 16-bit integers. You cannot pass NULL for
	any of the channles!
	
	\retval true Success.
	\retval false The display driver or the hardware does not support gamma translation.
*/
bool 
MgVideoSystem::SetGammaRamp(Uint16 * red, Uint16 * green, Uint16 * blue)
{
	return (SDL_SetGammaRamp(red, green, blue) == 0) ? true : false;
}

/*! Each table is an array of 256 16-bit integers. Any of the pointers may be NULL
	to ignore that channel.
*/
bool 
MgVideoSystem::GetGammaRamp(Uint16 * red, Uint16 * green, Uint16 * blue)
{
	return (SDL_GetGammaRamp(red, green, blue) == 0) ? true : false;
}

/*! \retval true Success.
	\retval false Error.
*/
bool 
MgVideoSystem::IconifyWindow()
{
	return (SDL_WM_IconifyWindow() == 1) ? true : false;
}

/*! If the display surface does not require lockign before accessing the pixel information,
	then the memory pointers will not change.
	
	\retval true Success.
	\retval false Error.
*/
bool 
MgVideoSystem::ToggleFullScreen()
{
	if (fGL || !fScreen)
		return false;
	return (SDL_WM_ToggleFullScreen(fScreen->fSurface) == 1) ? true : false;
}

MgGrabMode 
MgVideoSystem::GrabInput(MgGrabMode mode)
{
	return SDL_WM_GrabInput(mode);
}

