// MgVideoSystem.h
//-----------------------------------------------------------------------------
/*
    libMgTk - Mage Toolkit
    Copyright (C) 2001 Vitaliy Mikitchenko (Mage Productions)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Vitaliy Mikitchenko
    webmaster@magep.cjb.net
*/

#ifndef MGVIDEOSYSTEM_H
#define MGVIDEOSYSTEM_H

#include <SDL_video.h>

#include "MgSubSystem.h"

#include "MgBeginCode.h"

/*! \file MgVideoSystem.h
	Contains the MgVideoSystem subsystem class that handles video
	initialization.
*/

/*
	These flags can be passed as the 'flags' parameter of the
	constructos in MgPainter (as well as MgSurface). They
	are equivalent to SDL's own flags, but are here so you
	won't have to use SDL's flags.
*/

//@{
//!	The following can be used in the second and third constructors of	MgPainter.
#define MG_HARDWARE_SURFACE SDL_HWSURFACE	/* Make it a hardware surface. */
#define MG_SOFTWARE_SURFACE SDL_SWSURFACE	/* Make it a software surface. */
#define MG_ASYNC_BLIT SDL_ASYNCBLIT			/* Blit this surface async. */
//@}

//@{
//! Can only be used in MgVideoSystem::SetVideoMode() and MgVideoSystem::SetGLVideoMode().
#define MG_ANY_FORMAT SDL_ANYFORMAT			/* Can be any pixel format. */
#define MG_HARDWARE_PALETTE SDL_HWPALETTE	/* Use hardware palette. */
#define MG_DOUBLE_BUFFER SDL_DOUBLEBUF		/* Try to double buffer. */
#define MG_FULLSCREEN SDL_FULLSCREEN		/* Make it fullscreen. */
#define MG_RESIZABLE SDL_RESIZABLE			/* The window is resizable. */
#define MG_NO_FRAME SDL_NOFRAME				/* No frame around window. */
//@}

//@{
//! Can only be used in MgVideoSystem::SetGLVideoMode().
#define MG_OPENGL SDL_OPENGL				/* Use OpenGL. You don't have to pass to SetGLVideoMode() as it is set by default. */
#define MG_OPENGELBLIT SDL_OPENGLBLIT		/* Use OGL blit. */
//@}

//@{
//!	Read-only (used by SDL internally).
#define MG_HARDWARE_ACCEL SDL_HWACCEL
#define MG_SRC_COLORKEY SDL_SRCCOLORKEY
#define MG_RLE_ACCEL_OK SDL_RLEACCELOK
#define MG_RLE_ACCEL SDL_RLEACCEL
#define MG_SRC_ALPHA SDL_SRCALPHA
#define MG_PRE_ALLOC SDL_PREALLOC
//@}

class MgScreen;
class MgGLScreen;
class MgPixelFormat;

/*!
	The video info class. You never construct this class yourself.
*/
class MGDLL MgVideoInfo
{
public:
	//! Constructor.
	MgVideoInfo(const SDL_VideoInfo * info);
	//! Destructor.
	~MgVideoInfo();
	
	//! Can you create hardware surfaces?
	bool HWSurface() const;
	//! Can you talk to a window manager?
	bool WindowManager() const;
	
	//! Are HW to HW blits accelerated?
	bool HWBlit() const;
	//! Are blits with colorkey accelerated?
	bool HWColorKeyBlit() const;
	//! Are blits with alpha accelerated?
	bool HWAlphaBlit() const;
	//! Are SW to HW blits accelerated?
	bool SWBlit() const;
	//! Are blits with colorkey accelerated?
	bool SWColorKeyBlit() const;
	//! Are blits with alpha accelerated?
	bool SWAlphaBlit() const;
	//! Is colorfill accelerated?
	bool FillBlit() const;
	
	//! Amount of video memory in KB
	Uint32 VideoMemoryKB() const;
	//! Amount of video memory in MB
	Uint32 VideoMemoryMB() const;
	
	//! Pixel format of the video surface.
	const MgPixelFormat * Format() const;

private:
	const SDL_VideoInfo * fInfo;
	MgPixelFormat * fFormat;
};

//! Define MgGrabMode. It is the same as SDL_GrabMode.
#define MgGrabMode SDL_GrabMode

/*!
	This is the video subsystem class.
*/
class MGDLL MgVideoSystem : public MgSubSystem
{
public:
	//! Constructor. Does nothing.
	MgVideoSystem(const char * title);
	//! Destructor. Does nothing.
	virtual ~MgVideoSystem();
	
	virtual bool IsRunning() const;

	virtual bool Init();
	
	virtual void Quit();

	//! Set the video mode.
	MgScreen * SetVideoMode(int width, int height, int depth, Uint32 flags);
	
	//! Set the OpenGL video mode.
	MgGLScreen * SetGLVideoMode(int width, int height, int depth, Uint32 flags);
	
	//! Get some info about the video hardware.
	const MgVideoInfo * GetVideoInfo();
	
	//! Check to see if the video mode is OK.
	int VideoModeOK(int width, int height, int bpp, Uint32 flags);
	
	//@{
	/*! Returns a pointer to an array of available screen dimensions for the
		given format and video flags, sorted largest to smallest.
		
		\param format Format to use. If NULL, GetVideoInfo()->Format() will be used.
		\param flags MG_* flags used in SetVideoMode()
		
		\retval NULL No dimensions available for the format.
		\retval (SDL_Rect**)-1 Any dimension is okay for the format.
	*/
	SDL_Rect ** ListModes(MgPixelFormat * format, Uint32 flags);
	SDL_Rect ** ListModes(SDL_PixelFormat * format, Uint32 flags);
	//@}
	
	//! Get the current window manager title.
	const char * Title() const;
	//! Set the window manager title.
	void SetTitle(const char * title);
	
	//! Set the gamma correctin for each of the color channels.
	bool SetGamma(float red, float green, float blue);
	//! Set the gamma translation table for red, green and blue channels.
	bool SetGammaRamp(Uint16 * red, Uint16 * green, Uint16 * blue);
	//! Get the current values for the gamma translation tables.
	bool GetGammaRamp(Uint16 * red, Uint16 * green, Uint16 * blue);
	
	//! Iconify the window.
	bool IconifyWindow();
	//! Toggle fullscreen mode without changing the contents of the screen.
	bool ToggleFullScreen();
	/*! From the SDL_video.h header:
	   Grabbing means that the mouse is confined to the application window,
	   and nearly all keyboard input is passed directly to the application,
	   and not interpreted by a window manager, if any.
	*/
	MgGrabMode GrabInput(MgGrabMode mode);

private:
	MgScreen * fScreen;
	MgGLScreen * fGLScreen;
	MgVideoInfo * fVideoInfo;
	char * fTitle;
	
	Uint32 fGL : 1;		/* set to true if using OpenGL */
};

#include "MgEndCode.h"

#endif
